from collections import defaultdict
from decimal import Decimal

from ..transform import moneda_decimales
from ..transform.helpers import strcode
from ..utils import iterate


def rounder(moneda):
    decimals = moneda_decimales(moneda)
    return lambda v: round(v, decimals)


def aggregate(sequence, keys: tuple, values: tuple, project=None) -> list:
    _res = defaultdict(lambda: tuple(None for _ in values))

    _keys = lambda obj: tuple(obj.get(i) for i in keys)
    _values = lambda obj: tuple(obj.get(i) for i in values)
    if callable(project):
        _project = tuple(project(i) for i in keys + values)
    else:
        _project = project or keys + values

    def add(t):
        a, b = t
        if a is None:
            return b
        if b is None:
            return a
        return a + b

    for dic in sequence:
        k = _keys(dic)
        _res[k] = tuple(add(t) for t in zip(_res[k], _values(dic)))

    return [
        dict(zip(_project, k + v))
        for k, v in _res.items()
    ]


def encode_impuesto(impuesto, tipo_factor, tasa_cuota: Decimal = None):
    impuesto = strcode(impuesto)
    if tipo_factor:
        impuesto += "|" + strcode(tipo_factor)
        if tasa_cuota is not None:
            impuesto += "|" + tasa_cuota.__format__(".6f")
    return impuesto


def make_impuesto(impuesto: dict, base, rnd_fn):
    _impuesto = impuesto["Impuesto"]
    tipo_factor = impuesto['TipoFactor']
    tasa_cuota = impuesto['TasaOCuota']
    if impuesto.get('Base') is not None:
        base = impuesto['Base']

    if impuesto.get('Importe') is not None:
        importe = impuesto['Importe']
    else:
        match tipo_factor:
            case "Tasa":
                importe = rnd_fn(base * tasa_cuota)
            case "Cuota":
                importe = tasa_cuota
            case "Exento":
                importe = None
            case _:
                raise ValueError("Invalid TipoFactor", tipo_factor)

    return {
        'Base': rnd_fn(base),
        'Impuesto': _impuesto,
        'TipoFactor': tipo_factor,
        'TasaOCuota': tasa_cuota,
        'Importe': importe
    }


def group_impuestos(elements, pfx="", ofx=""):
    retenciones = aggregate(
        (t for c in iterate(elements) for t in iterate((c[f"Impuestos{pfx}"] or {}).get(f"Retenciones{pfx}"))),
        keys=(f"Impuesto{pfx}",),
        values=(f"Importe{pfx}",),
        project=lambda i: i[:len(i) - len(pfx)] + ofx,
    )
    traslados = aggregate(
        (t for c in iterate(elements) for t in iterate((c[f"Impuestos{pfx}"] or {}).get(f"Traslados{pfx}"))),
        keys=(f"Impuesto{pfx}", f"TipoFactor{pfx}", f"TasaOCuota{pfx}"),
        values=(f"Base{pfx}", f"Importe{pfx}"),
        project=lambda i: i[:len(i) - len(pfx)] + ofx,
    )

    impuestos = {}
    if retenciones:
        impuestos[f"Retenciones{ofx}"] = retenciones

    if traslados:
        impuestos[f"Traslados{ofx}"] = traslados

    return impuestos


def make_impuestos(conceptos):
    impuestos = group_impuestos(conceptos)

    if retenciones := impuestos.get('Retenciones'):
        impuestos['Retenciones'] = retenciones
        impuestos['TotalImpuestosRetenidos'] = sum(i["Importe"] for i in retenciones if i["Importe"])

    if traslados := impuestos.get('Traslados'):
        impuestos['Traslados'] = traslados
        impuestos['TotalImpuestosTrasladados'] = sum(i["Importe"] for i in traslados if i["Importe"])

    return impuestos or None


def make_impuestos_dr(conceptos):
    impuestos = {}
    for imp_t in ("Retenciones", "Traslados"):
        imp = aggregate(
            (t for c in conceptos for t in iterate((c["Impuestos"] or {}).get(imp_t))),
            keys=("Impuesto", 'TipoFactor', "TasaOCuota"),
            values=("Base", "Importe"),
            project=lambda i: i + "DR",
        )

        if imp:
            impuestos[imp_t + 'DR'] = imp

    return impuestos or None


def calculate_partial(value, imp_saldo_ant, imp_pagado, total, rnd_fn):
    if value is None:
        return None

    p_tot_pagado = (total - imp_saldo_ant + imp_pagado) / total
    p_anterior = (total - imp_saldo_ant) / total

    return rnd_fn(value * p_tot_pagado) - rnd_fn(value * p_anterior)


def make_impuestos_dr_parcial(conceptos, imp_saldo_ant: Decimal, imp_pagado: Decimal, total: Decimal, rnd_fn):
    p_tot_pagado = (total - imp_saldo_ant + imp_pagado) / total
    p_anterior = (total - imp_saldo_ant) / total

    impuestos = make_impuestos_dr(conceptos)
    for imp_t in ("RetencionesDR", "TrasladosDR"):
        for t in impuestos.get(imp_t, []):
            if t["ImporteDR"] is not None:
                t["ImporteDR"] = rnd_fn(t["ImporteDR"] * p_tot_pagado) - rnd_fn(t["ImporteDR"] * p_anterior)
            t["BaseDR"] = rnd_fn(t["BaseDR"] * p_tot_pagado) - rnd_fn(t["BaseDR"] * p_anterior)

    return impuestos


def make_impuestos_p(documentos):
    impuestos = group_impuestos(documentos, pfx="DR", ofx="P")
    return impuestos or None


RETENCIONES_MAP = {
    '001': 'TotalRetencionesISR',
    "002": 'TotalRetencionesIVA',
    '003': 'TotalRetencionesIEPS'
}


def make_pago_totales(pagos):
    impuestos = defaultdict(lambda: 0)

    for p in iterate(pagos):
        tipo_cambio = p.get('TipoCambioP', 1)
        impuestos['MontoTotalPagos'] += p['Monto'] * tipo_cambio

        for retencion in iterate((p["ImpuestosP"] or {}).get("RetencionesP")):
            impuestos[RETENCIONES_MAP[retencion["ImpuestoP"]]] += retencion["ImporteP"] * tipo_cambio

        for traslado in iterate((p["ImpuestosP"] or {}).get("TrasladosP")):
            match (traslado["ImpuestoP"], traslado["TipoFactorP"], str(traslado["TasaOCuotaP"])):
                case ("002", "Tasa", "0.160000"):
                    impuestos['TotalTrasladosBaseIVA16'] += traslado["BaseP"] * tipo_cambio
                    impuestos['TotalTrasladosImpuestoIVA16'] += traslado["ImporteP"] * tipo_cambio
                case ("002", "Tasa", "0.080000"):
                    impuestos['TotalTrasladosBaseIVA8'] += traslado["BaseP"] * tipo_cambio
                    impuestos['TotalTrasladosImpuestoIVA8'] += traslado["ImporteP"] * tipo_cambio
                case ("002", "Tasa", "0.000000"):
                    impuestos['TotalTrasladosBaseIVA0'] += traslado["BaseP"] * tipo_cambio
                    impuestos['TotalTrasladosImpuestoIVA0'] += traslado["ImporteP"] * tipo_cambio
                case ("002", "Tasa", _):
                    raise ValueError("Unknown Impuesto")
                case ("002", "Exento", _):
                    impuestos['TotalTrasladosBaseIVAExento'] += traslado["BaseP"] * tipo_cambio

    return impuestos



from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class ActivoVirtualType(XElement):
    def __init__(
            self,
            tipo_activo_virtual: int,
            cantidad_activo_virtual: str,
            descripcion_activo_virtual: str = None,
    ): 
        """
        
        :param tipo_activo_virtual:
        :param cantidad_activo_virtual:
        :param descripcion_activo_virtual:
        """
        
        super().__init__({
            'TipoActivoVirtual': tipo_activo_virtual,
            'CantidadActivoVirtual': cantidad_activo_virtual,
            'DescripcionActivoVirtual': descripcion_activo_virtual,
        })
        

class DatosOperacionFinancieraType(XElement):
    def __init__(
            self,
            instrumento_monetario: str,
            monto_operacion: str,
            fecha_pago: date = None,
            activo_virtual: ActivoVirtualType | dict = None,
            moneda: str = None,
    ): 
        """
        
        :param instrumento_monetario:
        :param monto_operacion:
        :param fecha_pago:
        :param activo_virtual:
        :param moneda:
        """
        
        super().__init__({
            'InstrumentoMonetario': instrumento_monetario,
            'MontoOperacion': monto_operacion,
            'FechaPago': fecha_pago,
            'ActivoVirtual': activo_virtual,
            'Moneda': moneda,
        })
        

class DatosSociedadMercantilType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            giro_mercantil: str,
            pais_nacionalidad: str,
            acciones_adquiridas: str,
            acciones_totales: str,
            datos_contraparte: TipoPersonaT1Type | dict,
            fecha_constitucion: date = None,
            rfc: str = None,
            folio_mercantil: str = None,
    ): 
        """
        
        :param denominacion_razon:
        :param giro_mercantil:
        :param pais_nacionalidad:
        :param acciones_adquiridas:
        :param acciones_totales:
        :param datos_contraparte:
        :param fecha_constitucion:
        :param rfc:
        :param folio_mercantil:
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'GiroMercantil': giro_mercantil,
            'PaisNacionalidad': pais_nacionalidad,
            'AccionesAdquiridas': acciones_adquiridas,
            'AccionesTotales': acciones_totales,
            'DatosContraparte': datos_contraparte,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
            'FolioMercantil': folio_mercantil,
        })
        

class CompraVentaEntidadesMercantilesType(XElement):
    def __init__(
            self,
            tipo_operacion: str,
            datos_sociedad_mercantil: DatosSociedadMercantilType | dict | Sequence[DatosSociedadMercantilType | dict],
    ): 
        """
        
        :param tipo_operacion:
        :param datos_sociedad_mercantil:
        """
        
        super().__init__({
            'TipoOperacion': tipo_operacion,
            'DatosSociedadMercantil': datos_sociedad_mercantil,
        })
        

class DatosComiteTecnicoType(XElement):
    def __init__(
            self,
            comite_tecnico: str,
    ): 
        """
        
        :param comite_tecnico:
        """
        
        super().__init__({
            'ComiteTecnico': comite_tecnico,
        })
        

class DatosFideicomisarioType(XElement):
    def __init__(
            self,
            datos_fideicomisarios_determinados: str,
            tipo_persona: TipoPersonaT1Type | dict = None,
    ): 
        """
        
        :param datos_fideicomisarios_determinados:
        :param tipo_persona:
        """
        
        super().__init__({
            'DatosFideicomisariosDeterminados': datos_fideicomisarios_determinados,
            'TipoPersona': tipo_persona,
        })
        

class PatrimonioOtroBienType(XElement):
    def __init__(
            self,
            descripcion: str,
            valor_bien: str,
    ): 
        """
        
        :param descripcion:
        :param valor_bien:
        """
        
        super().__init__({
            'Descripcion': descripcion,
            'ValorBien': valor_bien,
        })
        

class PatrimonioInmuebleType(XElement):
    def __init__(
            self,
            tipo_inmueble: str,
            codigo_postal: str,
            folio_real: str,
            importe_garantia: str,
    ): 
        """
        
        :param tipo_inmueble:
        :param codigo_postal:
        :param folio_real:
        :param importe_garantia:
        """
        
        super().__init__({
            'TipoInmueble': tipo_inmueble,
            'CodigoPostal': codigo_postal,
            'FolioReal': folio_real,
            'ImporteGarantia': importe_garantia,
        })
        

class PatrimonioMonetarioType(XElement):
    def __init__(
            self,
            moneda: str,
            monto_operacion: str,
    ): 
        """
        
        :param moneda:
        :param monto_operacion:
        """
        
        super().__init__({
            'Moneda': moneda,
            'MontoOperacion': monto_operacion,
        })
        

class DatosTipoPatrimonioType(XElement):
    def __init__(
            self,
            patrimonio_monetario: PatrimonioMonetarioType | dict = None,
            patrimonio_inmueble: PatrimonioInmuebleType | dict = None,
            patrimonio_otro_bien: PatrimonioOtroBienType | dict = None,
    ): 
        """
        
        :param patrimonio_monetario:
        :param patrimonio_inmueble:
        :param patrimonio_otro_bien:
        """
        
        super().__init__({
            'PatrimonioMonetario': patrimonio_monetario,
            'PatrimonioInmueble': patrimonio_inmueble,
            'PatrimonioOtroBien': patrimonio_otro_bien,
        })
        

class DatosFideicomitenteType(XElement):
    def __init__(
            self,
            tipo_persona: TipoPersonaT1Type | dict,
            datos_tipo_patrimonio: DatosTipoPatrimonioType | dict | Sequence[DatosTipoPatrimonioType | dict],
    ): 
        """
        
        :param tipo_persona:
        :param datos_tipo_patrimonio:
        """
        
        super().__init__({
            'TipoPersona': tipo_persona,
            'DatosTipoPatrimonio': datos_tipo_patrimonio,
        })
        

class ConstitucionFideicomisoType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            objeto_fideicomiso: str,
            monto_total_patrimonio: str,
            datos_fideicomitente: DatosFideicomitenteType | dict | Sequence[DatosFideicomitenteType | dict],
            datos_fideicomisario: DatosFideicomisarioType | dict | Sequence[DatosFideicomisarioType | dict],
            datos_miembro_comite_tecnico: DatosComiteTecnicoType | dict,
            rfc: str = None,
            identificador_fideicomiso: str = None,
    ): 
        """
        
        :param denominacion_razon:
        :param objeto_fideicomiso:
        :param monto_total_patrimonio:
        :param datos_fideicomitente:
        :param datos_fideicomisario:
        :param datos_miembro_comite_tecnico:
        :param rfc:
        :param identificador_fideicomiso:
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'ObjetoFideicomiso': objeto_fideicomiso,
            'MontoTotalPatrimonio': monto_total_patrimonio,
            'DatosFideicomitente': datos_fideicomitente,
            'DatosFideicomisario': datos_fideicomisario,
            'DatosMiembroComiteTecnico': datos_miembro_comite_tecnico,
            'Rfc': rfc,
            'IdentificadorFideicomiso': identificador_fideicomiso,
        })
        

class TipoPersonaMoralType(XElement):
    def __init__(
            self,
            persona_moral: PersonaMoralSimpleType | dict = None,
            fideicomiso: FideicomisoSimpleType | dict = None,
    ): 
        """
        
        :param persona_moral:
        :param fideicomiso:
        """
        
        super().__init__({
            'PersonaMoral': persona_moral,
            'Fideicomiso': fideicomiso,
        })
        

class AdmonPersonasMoralesType(XElement):
    def __init__(
            self,
            tipo_administracion: str,
            tipo_operacion: str,
            persona_moral_aviso: str,
            tipo_persona: TipoPersonaMoralType | dict = None,
    ): 
        """
        
        :param tipo_administracion:
        :param tipo_operacion:
        :param persona_moral_aviso:
        :param tipo_persona:
        """
        
        super().__init__({
            'TipoAdministracion': tipo_administracion,
            'TipoOperacion': tipo_operacion,
            'PersonaMoralAviso': persona_moral_aviso,
            'TipoPersona': tipo_persona,
        })
        

class EscindidaType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            giro_mercantil: str,
            capital_social_fijo: str,
            numero_total_acciones: str,
            datos_accionista: DatosAccionistaT2Type | dict | Sequence[DatosAccionistaT2Type | dict],
            fecha_constitucion: date = None,
            rfc: str = None,
            capital_social_variable: str = None,
            folio_mercantil: str = None,
    ): 
        """
        
        :param denominacion_razon:
        :param pais_nacionalidad:
        :param giro_mercantil:
        :param capital_social_fijo:
        :param numero_total_acciones:
        :param datos_accionista:
        :param fecha_constitucion:
        :param rfc:
        :param capital_social_variable:
        :param folio_mercantil:
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'GiroMercantil': giro_mercantil,
            'CapitalSocialFijo': capital_social_fijo,
            'NumeroTotalAcciones': numero_total_acciones,
            'DatosAccionista': datos_accionista,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
            'CapitalSocialVariable': capital_social_variable,
            'FolioMercantil': folio_mercantil,
        })
        

class DatosEscindidasType(XElement):
    def __init__(
            self,
            escindidas_determinadas: str,
            dato_escindida: EscindidaType | dict | Sequence[EscindidaType | dict] = None,
    ): 
        """
        
        :param escindidas_determinadas:
        :param dato_escindida:
        """
        
        super().__init__({
            'EscindidasDeterminadas': escindidas_determinadas,
            'DatoEscindida': dato_escindida,
        })
        

class DatosAccionistaT3Type(XElement):
    def __init__(
            self,
            tipo_persona: TipoPersonaT1Type | dict,
            numero_acciones: str,
    ): 
        """
        
        :param tipo_persona:
        :param numero_acciones:
        """
        
        super().__init__({
            'TipoPersona': tipo_persona,
            'NumeroAcciones': numero_acciones,
        })
        

class DatosEscindenteType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            giro_mercantil: str,
            capital_social_fijo: str,
            escindente_subsiste: str,
            fecha_constitucion: date = None,
            rfc: str = None,
            capital_social_variable: str = None,
            folio_mercantil: str = None,
            datos_accionista_escindente: DatosAccionistaT3Type | dict | Sequence[DatosAccionistaT3Type | dict] = None,
    ): 
        """
        
        :param denominacion_razon:
        :param pais_nacionalidad:
        :param giro_mercantil:
        :param capital_social_fijo:
        :param escindente_subsiste:
        :param fecha_constitucion:
        :param rfc:
        :param capital_social_variable:
        :param folio_mercantil:
        :param datos_accionista_escindente:
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'GiroMercantil': giro_mercantil,
            'CapitalSocialFijo': capital_social_fijo,
            'EscindenteSubsiste': escindente_subsiste,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
            'CapitalSocialVariable': capital_social_variable,
            'FolioMercantil': folio_mercantil,
            'DatosAccionistaEscindente': datos_accionista_escindente,
        })
        

class EscisionType(XElement):
    def __init__(
            self,
            datos_escindente: DatosEscindenteType | dict,
            datos_escindidas: DatosEscindidasType | dict,
    ): 
        """
        
        :param datos_escindente:
        :param datos_escindidas:
        """
        
        super().__init__({
            'DatosEscindente': datos_escindente,
            'DatosEscindidas': datos_escindidas,
        })
        

class DatosAccionistaT2Type(XElement):
    def __init__(
            self,
            tipo_persona: TipoPersonaT1Type | dict,
            numero_acciones: str,
    ): 
        """
        
        :param tipo_persona:
        :param numero_acciones:
        """
        
        super().__init__({
            'TipoPersona': tipo_persona,
            'NumeroAcciones': numero_acciones,
        })
        

class FusionanteType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            giro_mercantil: str,
            capital_social_fijo: str,
            numero_total_acciones: str,
            datos_accionista: DatosAccionistaT2Type | dict | Sequence[DatosAccionistaT2Type | dict],
            fecha_constitucion: date = None,
            rfc: str = None,
            capital_social_variable: str = None,
            folio_mercantil: str = None,
    ): 
        """
        
        :param denominacion_razon:
        :param pais_nacionalidad:
        :param giro_mercantil:
        :param capital_social_fijo:
        :param numero_total_acciones:
        :param datos_accionista:
        :param fecha_constitucion:
        :param rfc:
        :param capital_social_variable:
        :param folio_mercantil:
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'GiroMercantil': giro_mercantil,
            'CapitalSocialFijo': capital_social_fijo,
            'NumeroTotalAcciones': numero_total_acciones,
            'DatosAccionista': datos_accionista,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
            'CapitalSocialVariable': capital_social_variable,
            'FolioMercantil': folio_mercantil,
        })
        

class DatosFusionanteType(XElement):
    def __init__(
            self,
            fusionante_determinadas: str,
            fusionante: FusionanteType | dict = None,
    ): 
        """
        
        :param fusionante_determinadas:
        :param fusionante:
        """
        
        super().__init__({
            'FusionanteDeterminadas': fusionante_determinadas,
            'Fusionante': fusionante,
        })
        

class DatosFusionadaType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            giro_mercantil: str,
            capital_social_fijo: str,
            fecha_constitucion: date = None,
            rfc: str = None,
            capital_social_variable: str = None,
            folio_mercantil: str = None,
    ): 
        """
        
        :param denominacion_razon:
        :param pais_nacionalidad:
        :param giro_mercantil:
        :param capital_social_fijo:
        :param fecha_constitucion:
        :param rfc:
        :param capital_social_variable:
        :param folio_mercantil:
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'GiroMercantil': giro_mercantil,
            'CapitalSocialFijo': capital_social_fijo,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
            'CapitalSocialVariable': capital_social_variable,
            'FolioMercantil': folio_mercantil,
        })
        

class DatosFusionadasType(XElement):
    def __init__(
            self,
            datos_fusionada: DatosFusionadaType | dict | Sequence[DatosFusionadaType | dict],
    ): 
        """
        
        :param datos_fusionada:
        """
        
        super().__init__({
            'DatosFusionada': datos_fusionada,
        })
        

class FusionType(XElement):
    def __init__(
            self,
            tipo_fusion: str,
            datos_fusionadas: DatosFusionadasType | dict,
            datos_fusionante: DatosFusionanteType | dict,
    ): 
        """
        
        :param tipo_fusion:
        :param datos_fusionadas:
        :param datos_fusionante:
        """
        
        super().__init__({
            'TipoFusion': tipo_fusion,
            'DatosFusionadas': datos_fusionadas,
            'DatosFusionante': datos_fusionante,
        })
        

class AportacionOtroBienType(XElement):
    def __init__(
            self,
            descripcion: str,
            valor_aportacion: str,
    ): 
        """
        
        :param descripcion:
        :param valor_aportacion:
        """
        
        super().__init__({
            'Descripcion': descripcion,
            'ValorAportacion': valor_aportacion,
        })
        

class AportacionInmuebleType(XElement):
    def __init__(
            self,
            tipo_inmueble: str,
            codigo_postal: str,
            folio_real: str,
            valor_aportacion: str,
    ): 
        """
        
        :param tipo_inmueble:
        :param codigo_postal:
        :param folio_real:
        :param valor_aportacion:
        """
        
        super().__init__({
            'TipoInmueble': tipo_inmueble,
            'CodigoPostal': codigo_postal,
            'FolioReal': folio_real,
            'ValorAportacion': valor_aportacion,
        })
        

class AportacionMonetariaType(XElement):
    def __init__(
            self,
            instrumento_monetario: str,
            moneda: str,
            monto_operacion: str,
    ): 
        """
        
        :param instrumento_monetario:
        :param moneda:
        :param monto_operacion:
        """
        
        super().__init__({
            'InstrumentoMonetario': instrumento_monetario,
            'Moneda': moneda,
            'MontoOperacion': monto_operacion,
        })
        

class DatosTipoAportacionType(XElement):
    def __init__(
            self,
            aportacion_monetaria: AportacionMonetariaType | dict = None,
            aportacion_inmueble: AportacionInmuebleType | dict = None,
            aportacion_otro_bien: AportacionOtroBienType | dict = None,
    ): 
        """
        
        :param aportacion_monetaria:
        :param aportacion_inmueble:
        :param aportacion_otro_bien:
        """
        
        super().__init__({
            'AportacionMonetaria': aportacion_monetaria,
            'AportacionInmueble': aportacion_inmueble,
            'AportacionOtroBien': aportacion_otro_bien,
        })
        

class PersonaMoralT2Type(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            giro_mercantil: str,
            fecha_constitucion: date = None,
            rfc: str = None,
    ): 
        """
        
        :param denominacion_razon:
        :param pais_nacionalidad:
        :param giro_mercantil:
        :param fecha_constitucion:
        :param rfc:
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'GiroMercantil': giro_mercantil,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
        })
        

class PersonaFisicaT2Type(XElement):
    def __init__(
            self,
            nombre: str,
            apellido_paterno: str,
            apellido_materno: str,
            pais_nacionalidad: str,
            actividad_economica: str,
            fecha_nacimiento: date = None,
            rfc: str = None,
            curp: str = None,
    ): 
        """
        
        :param nombre:
        :param apellido_paterno:
        :param apellido_materno:
        :param pais_nacionalidad:
        :param actividad_economica:
        :param fecha_nacimiento:
        :param rfc:
        :param curp:
        """
        
        super().__init__({
            'Nombre': nombre,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'PaisNacionalidad': pais_nacionalidad,
            'ActividadEconomica': actividad_economica,
            'FechaNacimiento': fecha_nacimiento,
            'Rfc': rfc,
            'Curp': curp,
        })
        

class TipoPersonaT2Type(XElement):
    def __init__(
            self,
            persona_fisica: PersonaFisicaT2Type | dict = None,
            persona_moral: PersonaMoralT2Type | dict = None,
            fideicomiso: FideicomisoSimpleType | dict = None,
    ): 
        """
        
        :param persona_fisica:
        :param persona_moral:
        :param fideicomiso:
        """
        
        super().__init__({
            'PersonaFisica': persona_fisica,
            'PersonaMoral': persona_moral,
            'Fideicomiso': fideicomiso,
        })
        

class DatosAportacionType(XElement):
    def __init__(
            self,
            datos_persona_aporta: TipoPersonaT2Type | dict,
            datos_tipo_aportacion: DatosTipoAportacionType | dict | Sequence[DatosTipoAportacionType | dict],
    ): 
        """
        
        :param datos_persona_aporta:
        :param datos_tipo_aportacion:
        """
        
        super().__init__({
            'DatosPersonaAporta': datos_persona_aporta,
            'DatosTipoAportacion': datos_tipo_aportacion,
        })
        

class OrganizacionAportacionesType(XElement):
    def __init__(
            self,
            motivo_aportacion: str,
            datos_aportacion: DatosAportacionType | dict | Sequence[DatosAportacionType | dict],
    ): 
        """
        
        :param motivo_aportacion:
        :param datos_aportacion:
        """
        
        super().__init__({
            'MotivoAportacion': motivo_aportacion,
            'DatosAportacion': datos_aportacion,
        })
        

class CapitalSocialType(XElement):
    def __init__(
            self,
            capital_fijo: str,
            capital_variable: str = None,
    ): 
        """
        
        :param capital_fijo:
        :param capital_variable:
        """
        
        super().__init__({
            'CapitalFijo': capital_fijo,
            'CapitalVariable': capital_variable,
        })
        

class DatosAccionistaT1Type(XElement):
    def __init__(
            self,
            cargo_accionista: str,
            tipo_persona: TipoPersonaT1Type | dict,
            numero_acciones: str,
    ): 
        """
        
        :param cargo_accionista:
        :param tipo_persona:
        :param numero_acciones:
        """
        
        super().__init__({
            'CargoAccionista': cargo_accionista,
            'TipoPersona': tipo_persona,
            'NumeroAcciones': numero_acciones,
        })
        

class ConstitucionSociedadesMercantilesType(XElement):
    def __init__(
            self,
            tipo_persona_moral: int,
            denominacion_razon: str,
            giro_mercantil: str,
            numero_total_acciones: str,
            entidad_federativa: int,
            consejo_vigilancia: str,
            motivo_constitucion: str,
            datos_accionista: DatosAccionistaT1Type | dict | Sequence[DatosAccionistaT1Type | dict],
            capital_social: CapitalSocialType | dict,
            tipo_persona_moral_otra: str = None,
            folio_mercantil: str = None,
            instrumento_publico: str = None,
    ): 
        """
        
        :param tipo_persona_moral:
        :param denominacion_razon:
        :param giro_mercantil:
        :param numero_total_acciones:
        :param entidad_federativa:
        :param consejo_vigilancia:
        :param motivo_constitucion:
        :param datos_accionista:
        :param capital_social:
        :param tipo_persona_moral_otra:
        :param folio_mercantil:
        :param instrumento_publico:
        """
        
        super().__init__({
            'TipoPersonaMoral': tipo_persona_moral,
            'DenominacionRazon': denominacion_razon,
            'GiroMercantil': giro_mercantil,
            'NumeroTotalAcciones': numero_total_acciones,
            'EntidadFederativa': entidad_federativa,
            'ConsejoVigilancia': consejo_vigilancia,
            'MotivoConstitucion': motivo_constitucion,
            'DatosAccionista': datos_accionista,
            'CapitalSocial': capital_social,
            'TipoPersonaMoralOtra': tipo_persona_moral_otra,
            'FolioMercantil': folio_mercantil,
            'InstrumentoPublico': instrumento_publico,
        })
        

class ActivoOtrosType(XElement):
    def __init__(
            self,
            descripcion_activo_administrado: str,
    ): 
        """
        
        :param descripcion_activo_administrado:
        """
        
        super().__init__({
            'DescripcionActivoAdministrado': descripcion_activo_administrado,
        })
        

class ActivoAdministradoType(XElement):
    def __init__(
            self,
            tipo_activo_administrado: int,
            descripcion_otro_activo_administrado: str = None,
    ): 
        """
        
        :param tipo_activo_administrado:
        :param descripcion_otro_activo_administrado:
        """
        
        super().__init__({
            'TipoActivoAdministrado': tipo_activo_administrado,
            'DescripcionOtroActivoAdministrado': descripcion_otro_activo_administrado,
        })
        

class AreaServicioType(XElement):
    def __init__(
            self,
            tipo_area_servicio: int,
            descripcion_otro_area_servicio: str = None,
    ): 
        """
        
        :param tipo_area_servicio:
        :param descripcion_otro_area_servicio:
        """
        
        super().__init__({
            'TipoAreaServicio': tipo_area_servicio,
            'DescripcionOtroAreaServicio': descripcion_otro_area_servicio,
        })
        

class ActivoOutsourcingType(XElement):
    def __init__(
            self,
            area_servicio: AreaServicioType | dict,
            activo_administrado: ActivoAdministradoType | dict,
            numero_empleados: int,
    ): 
        """
        
        :param area_servicio:
        :param activo_administrado:
        :param numero_empleados:
        """
        
        super().__init__({
            'AreaServicio': area_servicio,
            'ActivoAdministrado': activo_administrado,
            'NumeroEmpleados': numero_empleados,
        })
        

class ActivoInmobiliarioType(XElement):
    def __init__(
            self,
            tipo_inmueble: str,
            valor_referencia: str,
            colonia: str,
            calle: str,
            numero_exterior: str,
            codigo_postal: str,
            folio_real: str,
            numero_interior: str = None,
    ): 
        """
        
        :param tipo_inmueble:
        :param valor_referencia:
        :param colonia:
        :param calle:
        :param numero_exterior:
        :param codigo_postal:
        :param folio_real:
        :param numero_interior:
        """
        
        super().__init__({
            'TipoInmueble': tipo_inmueble,
            'ValorReferencia': valor_referencia,
            'Colonia': colonia,
            'Calle': calle,
            'NumeroExterior': numero_exterior,
            'CodigoPostal': codigo_postal,
            'FolioReal': folio_real,
            'NumeroInterior': numero_interior,
        })
        

class ActivoBancoType(XElement):
    def __init__(
            self,
            estatus_manejo: str,
            clave_tipo_institucion: int,
            nombre_institucion: str,
            numero_cuenta: str,
    ): 
        """
        
        :param estatus_manejo:
        :param clave_tipo_institucion:
        :param nombre_institucion:
        :param numero_cuenta:
        """
        
        super().__init__({
            'EstatusManejo': estatus_manejo,
            'ClaveTipoInstitucion': clave_tipo_institucion,
            'NombreInstitucion': nombre_institucion,
            'NumeroCuenta': numero_cuenta,
        })
        

class TipoActivoType(XElement):
    def __init__(
            self,
            activo_banco: ActivoBancoType | dict = None,
            activo_inmobiliario: ActivoInmobiliarioType | dict = None,
            activo_outsourcing: ActivoOutsourcingType | dict = None,
            activo_otros: ActivoOtrosType | dict = None,
    ): 
        """
        
        :param activo_banco:
        :param activo_inmobiliario:
        :param activo_outsourcing:
        :param activo_otros:
        """
        
        super().__init__({
            'ActivoBanco': activo_banco,
            'ActivoInmobiliario': activo_inmobiliario,
            'ActivoOutsourcing': activo_outsourcing,
            'ActivoOtros': activo_otros,
        })
        

class AdministracionRecursosType(XElement):
    def __init__(
            self,
            tipo_activo: TipoActivoType | dict | Sequence[TipoActivoType | dict],
            numero_operaciones: int,
    ): 
        """
        
        :param tipo_activo:
        :param numero_operaciones:
        """
        
        super().__init__({
            'TipoActivo': tipo_activo,
            'NumeroOperaciones': numero_operaciones,
        })
        

class CaracteristicasInmueble2Type(XElement):
    def __init__(
            self,
            tipo_inmueble: str,
            valor_referencia: str,
            colonia: str,
            calle: str,
            numero_exterior: str,
            codigo_postal: str,
            dimension_terreno: str,
            dimension_construido: str,
            folio_real: str,
            numero_interior: str = None,
    ): 
        """
        
        :param tipo_inmueble:
        :param valor_referencia:
        :param colonia:
        :param calle:
        :param numero_exterior:
        :param codigo_postal:
        :param dimension_terreno:
        :param dimension_construido:
        :param folio_real:
        :param numero_interior:
        """
        
        super().__init__({
            'TipoInmueble': tipo_inmueble,
            'ValorReferencia': valor_referencia,
            'Colonia': colonia,
            'Calle': calle,
            'NumeroExterior': numero_exterior,
            'CodigoPostal': codigo_postal,
            'DimensionTerreno': dimension_terreno,
            'DimensionConstruido': dimension_construido,
            'FolioReal': folio_real,
            'NumeroInterior': numero_interior,
        })
        

class CesionDerechosInmueblesType(XElement):
    def __init__(
            self,
            figura_cliente: str,
            tipo_cesion: str,
            datos_contraparte: DatosContraparteType | dict | Sequence[DatosContraparteType | dict],
            caracteristicas_inmueble: CaracteristicasInmueble2Type | dict | Sequence[CaracteristicasInmueble2Type | dict],
    ): 
        """
        
        :param figura_cliente:
        :param tipo_cesion:
        :param datos_contraparte:
        :param caracteristicas_inmueble:
        """
        
        super().__init__({
            'FiguraCliente': figura_cliente,
            'TipoCesion': tipo_cesion,
            'DatosContraparte': datos_contraparte,
            'CaracteristicasInmueble': caracteristicas_inmueble,
        })
        

class ContratoType(XElement):
    def __init__(
            self,
            fecha_contrato: date,
            valor_referencia: str,
    ): 
        """
        
        :param fecha_contrato:
        :param valor_referencia:
        """
        
        super().__init__({
            'FechaContrato': fecha_contrato,
            'ValorReferencia': valor_referencia,
        })
        

class DatosInstrumentoType(XElement):
    def __init__(
            self,
            numero_instrumento_publico: str,
            fecha_instrumento_publico: date,
            notario_instrumento_publico: str,
            entidad_instrumento_publico: int,
            valor_referencia: str,
    ): 
        """
        
        :param numero_instrumento_publico:
        :param fecha_instrumento_publico:
        :param notario_instrumento_publico:
        :param entidad_instrumento_publico:
        :param valor_referencia:
        """
        
        super().__init__({
            'NumeroInstrumentoPublico': numero_instrumento_publico,
            'FechaInstrumentoPublico': fecha_instrumento_publico,
            'NotarioInstrumentoPublico': notario_instrumento_publico,
            'EntidadInstrumentoPublico': entidad_instrumento_publico,
            'ValorReferencia': valor_referencia,
        })
        

class ContratoInstrumentoType(XElement):
    def __init__(
            self,
            datos_instrumento_publico: DatosInstrumentoType | dict = None,
            contrato: ContratoType | dict = None,
    ): 
        """
        
        :param datos_instrumento_publico:
        :param contrato:
        """
        
        super().__init__({
            'DatosInstrumentoPublico': datos_instrumento_publico,
            'Contrato': contrato,
        })
        

class CaracteristicasInmueble1Type(XElement):
    def __init__(
            self,
            tipo_inmueble: str,
            colonia: str,
            calle: str,
            numero_exterior: str,
            codigo_postal: str,
            dimension_terreno: str,
            dimension_construido: str,
            folio_real: str,
            contrato_instrumento_publico: ContratoInstrumentoType | dict,
            numero_interior: str = None,
    ): 
        """
        
        :param tipo_inmueble:
        :param colonia:
        :param calle:
        :param numero_exterior:
        :param codigo_postal:
        :param dimension_terreno:
        :param dimension_construido:
        :param folio_real:
        :param contrato_instrumento_publico:
        :param numero_interior:
        """
        
        super().__init__({
            'TipoInmueble': tipo_inmueble,
            'Colonia': colonia,
            'Calle': calle,
            'NumeroExterior': numero_exterior,
            'CodigoPostal': codigo_postal,
            'DimensionTerreno': dimension_terreno,
            'DimensionConstruido': dimension_construido,
            'FolioReal': folio_real,
            'ContratoInstrumentoPublico': contrato_instrumento_publico,
            'NumeroInterior': numero_interior,
        })
        

class PersonaMoralT1Type(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            fecha_constitucion: date = None,
            rfc: str = None,
    ): 
        """
        
        :param denominacion_razon:
        :param pais_nacionalidad:
        :param fecha_constitucion:
        :param rfc:
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
        })
        

class PersonaFisicaT1Type(XElement):
    def __init__(
            self,
            nombre: str,
            apellido_paterno: str,
            apellido_materno: str,
            pais_nacionalidad: str,
            fecha_nacimiento: date = None,
            rfc: str = None,
            curp: str = None,
    ): 
        """
        
        :param nombre:
        :param apellido_paterno:
        :param apellido_materno:
        :param pais_nacionalidad:
        :param fecha_nacimiento:
        :param rfc:
        :param curp:
        """
        
        super().__init__({
            'Nombre': nombre,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'PaisNacionalidad': pais_nacionalidad,
            'FechaNacimiento': fecha_nacimiento,
            'Rfc': rfc,
            'Curp': curp,
        })
        

class TipoPersonaT1Type(XElement):
    def __init__(
            self,
            persona_fisica: PersonaFisicaT1Type | dict = None,
            persona_moral: PersonaMoralT1Type | dict = None,
            fideicomiso: FideicomisoSimpleType | dict = None,
    ): 
        """
        
        :param persona_fisica:
        :param persona_moral:
        :param fideicomiso:
        """
        
        super().__init__({
            'PersonaFisica': persona_fisica,
            'PersonaMoral': persona_moral,
            'Fideicomiso': fideicomiso,
        })
        

class DatosContraparteType(XElement):
    def __init__(
            self,
            tipo_persona: TipoPersonaT1Type | dict,
    ): 
        """
        
        :param tipo_persona:
        """
        
        super().__init__({
            'TipoPersona': tipo_persona,
        })
        

class CompraVentaInmueblesType(XElement):
    def __init__(
            self,
            tipo_operacion: str,
            valor_pactado: str,
            datos_contraparte: DatosContraparteType | dict | Sequence[DatosContraparteType | dict],
            caracteristicas_inmueble: CaracteristicasInmueble1Type | dict | Sequence[CaracteristicasInmueble1Type | dict],
    ): 
        """
        
        :param tipo_operacion:
        :param valor_pactado:
        :param datos_contraparte:
        :param caracteristicas_inmueble:
        """
        
        super().__init__({
            'TipoOperacion': tipo_operacion,
            'ValorPactado': valor_pactado,
            'DatosContraparte': datos_contraparte,
            'CaracteristicasInmueble': caracteristicas_inmueble,
        })
        

class TipoActividadType(XElement):
    def __init__(
            self,
            compra_venta_inmuebles: CompraVentaInmueblesType | dict = None,
            cesion_derechos_inmuebles: CesionDerechosInmueblesType | dict = None,
            administracion_recursos: AdministracionRecursosType | dict = None,
            constitucion_sociedades_mercantiles: ConstitucionSociedadesMercantilesType | dict = None,
            organizacion_aportaciones: OrganizacionAportacionesType | dict = None,
            fusion: FusionType | dict = None,
            escision: EscisionType | dict = None,
            administracion_personas_morales: AdmonPersonasMoralesType | dict = None,
            constitucion_fideicomiso: ConstitucionFideicomisoType | dict = None,
            compra_venta_entidades_mercantiles: CompraVentaEntidadesMercantilesType | dict = None,
    ): 
        """
        
        :param compra_venta_inmuebles:
        :param cesion_derechos_inmuebles:
        :param administracion_recursos:
        :param constitucion_sociedades_mercantiles:
        :param organizacion_aportaciones:
        :param fusion:
        :param escision:
        :param administracion_personas_morales:
        :param constitucion_fideicomiso:
        :param compra_venta_entidades_mercantiles:
        """
        
        super().__init__({
            'CompraVentaInmuebles': compra_venta_inmuebles,
            'CesionDerechosInmuebles': cesion_derechos_inmuebles,
            'AdministracionRecursos': administracion_recursos,
            'ConstitucionSociedadesMercantiles': constitucion_sociedades_mercantiles,
            'OrganizacionAportaciones': organizacion_aportaciones,
            'Fusion': fusion,
            'Escision': escision,
            'AdministracionPersonasMorales': administracion_personas_morales,
            'ConstitucionFideicomiso': constitucion_fideicomiso,
            'CompraVentaEntidadesMercantiles': compra_venta_entidades_mercantiles,
        })
        

class DatosOperacionType(XElement):
    def __init__(
            self,
            fecha_operacion: date,
            tipo_actividad: TipoActividadType | dict,
            datos_operacion_financiera: DatosOperacionFinancieraType | dict | Sequence[DatosOperacionFinancieraType | dict],
    ): 
        """
        
        :param fecha_operacion:
        :param tipo_actividad:
        :param datos_operacion_financiera:
        """
        
        super().__init__({
            'FechaOperacion': fecha_operacion,
            'TipoActividad': tipo_actividad,
            'DatosOperacionFinanciera': datos_operacion_financiera,
        })
        

class DetalleOperacionesType(XElement):
    def __init__(
            self,
            datos_operacion: DatosOperacionType | dict | Sequence[DatosOperacionType | dict],
    ): 
        """
        
        :param datos_operacion:
        """
        
        super().__init__({
            'DatosOperacion': datos_operacion,
        })
        

class FideicomisoSimpleType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            rfc: str = None,
            identificador_fideicomiso: str = None,
    ): 
        """
        
        :param denominacion_razon:
        :param rfc:
        :param identificador_fideicomiso:
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'Rfc': rfc,
            'IdentificadorFideicomiso': identificador_fideicomiso,
        })
        

class PersonaMoralSimpleType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            fecha_constitucion: date = None,
            rfc: str = None,
            pais_nacionalidad: str = None,
    ): 
        """
        
        :param denominacion_razon:
        :param fecha_constitucion:
        :param rfc:
        :param pais_nacionalidad:
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
            'PaisNacionalidad': pais_nacionalidad,
        })
        

class PersonaFisicaSimpleType(XElement):
    def __init__(
            self,
            nombre: str,
            apellido_paterno: str,
            apellido_materno: str,
            fecha_nacimiento: date = None,
            rfc: str = None,
            curp: str = None,
            pais_nacionalidad: str = None,
    ): 
        """
        
        :param nombre:
        :param apellido_paterno:
        :param apellido_materno:
        :param fecha_nacimiento:
        :param rfc:
        :param curp:
        :param pais_nacionalidad:
        """
        
        super().__init__({
            'Nombre': nombre,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'FechaNacimiento': fecha_nacimiento,
            'Rfc': rfc,
            'Curp': curp,
            'PaisNacionalidad': pais_nacionalidad,
        })
        

class TipoPersonaSimpleType(XElement):
    def __init__(
            self,
            persona_fisica: PersonaFisicaSimpleType | dict = None,
            persona_moral: PersonaMoralSimpleType | dict = None,
            fideicomiso: FideicomisoSimpleType | dict = None,
    ): 
        """
        
        :param persona_fisica:
        :param persona_moral:
        :param fideicomiso:
        """
        
        super().__init__({
            'PersonaFisica': persona_fisica,
            'PersonaMoral': persona_moral,
            'Fideicomiso': fideicomiso,
        })
        

class DuenoBeneficiarioType(XElement):
    def __init__(
            self,
            tipo_persona: TipoPersonaSimpleType | dict,
    ): 
        """
        
        :param tipo_persona:
        """
        
        super().__init__({
            'TipoPersona': tipo_persona,
        })
        

class TelefonoType(XElement):
    def __init__(
            self,
            clave_pais: str = None,
            numero_telefono: str = None,
            correo_electronico: str = None,
    ): 
        """
        
        :param clave_pais:
        :param numero_telefono:
        :param correo_electronico:
        """
        
        super().__init__({
            'ClavePais': clave_pais,
            'NumeroTelefono': numero_telefono,
            'CorreoElectronico': correo_electronico,
        })
        

class ExtranjeroType(XElement):
    def __init__(
            self,
            pais: str,
            estado_provincia: str,
            ciudad_poblacion: str,
            colonia: str,
            calle: str,
            numero_exterior: str,
            codigo_postal: str,
            numero_interior: str = None,
    ): 
        """
        
        :param pais:
        :param estado_provincia:
        :param ciudad_poblacion:
        :param colonia:
        :param calle:
        :param numero_exterior:
        :param codigo_postal:
        :param numero_interior:
        """
        
        super().__init__({
            'Pais': pais,
            'EstadoProvincia': estado_provincia,
            'CiudadPoblacion': ciudad_poblacion,
            'Colonia': colonia,
            'Calle': calle,
            'NumeroExterior': numero_exterior,
            'CodigoPostal': codigo_postal,
            'NumeroInterior': numero_interior,
        })
        

class NacionalType(XElement):
    def __init__(
            self,
            colonia: str,
            calle: str,
            numero_exterior: str,
            codigo_postal: str,
            numero_interior: str = None,
    ): 
        """
        
        :param colonia:
        :param calle:
        :param numero_exterior:
        :param codigo_postal:
        :param numero_interior:
        """
        
        super().__init__({
            'Colonia': colonia,
            'Calle': calle,
            'NumeroExterior': numero_exterior,
            'CodigoPostal': codigo_postal,
            'NumeroInterior': numero_interior,
        })
        

class TipoDomicilioType(XElement):
    def __init__(
            self,
            nacional: NacionalType | dict = None,
            extranjero: ExtranjeroType | dict = None,
    ): 
        """
        
        :param nacional:
        :param extranjero:
        """
        
        super().__init__({
            'Nacional': nacional,
            'Extranjero': extranjero,
        })
        

class RepresentanteApoderadoType(XElement):
    def __init__(
            self,
            nombre: str,
            apellido_paterno: str,
            apellido_materno: str,
            fecha_nacimiento: date = None,
            rfc: str = None,
            curp: str = None,
    ): 
        """
        
        :param nombre:
        :param apellido_paterno:
        :param apellido_materno:
        :param fecha_nacimiento:
        :param rfc:
        :param curp:
        """
        
        super().__init__({
            'Nombre': nombre,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'FechaNacimiento': fecha_nacimiento,
            'Rfc': rfc,
            'Curp': curp,
        })
        

class FideicomisoType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            apoderado_delegado: RepresentanteApoderadoType | dict | Sequence[RepresentanteApoderadoType | dict],
            rfc: str = None,
            identificador_fideicomiso: str = None,
    ): 
        """
        
        :param denominacion_razon:
        :param apoderado_delegado:
        :param rfc:
        :param identificador_fideicomiso:
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'ApoderadoDelegado': apoderado_delegado,
            'Rfc': rfc,
            'IdentificadorFideicomiso': identificador_fideicomiso,
        })
        

class PersonaMoralType(XElement):
    def __init__(
            self,
            denominacion_razon: str,
            pais_nacionalidad: str,
            giro_mercantil: str,
            representante_apoderado: RepresentanteApoderadoType | dict | Sequence[RepresentanteApoderadoType | dict],
            fecha_constitucion: date = None,
            rfc: str = None,
    ): 
        """
        
        :param denominacion_razon:
        :param pais_nacionalidad:
        :param giro_mercantil:
        :param representante_apoderado:
        :param fecha_constitucion:
        :param rfc:
        """
        
        super().__init__({
            'DenominacionRazon': denominacion_razon,
            'PaisNacionalidad': pais_nacionalidad,
            'GiroMercantil': giro_mercantil,
            'RepresentanteApoderado': representante_apoderado,
            'FechaConstitucion': fecha_constitucion,
            'Rfc': rfc,
        })
        

class PersonaFisicaType(XElement):
    def __init__(
            self,
            nombre: str,
            apellido_paterno: str,
            apellido_materno: str,
            pais_nacionalidad: str,
            actividad_economica: str,
            fecha_nacimiento: date = None,
            rfc: str = None,
            curp: str = None,
            representante_apoderado: RepresentanteApoderadoType | dict | Sequence[RepresentanteApoderadoType | dict] = None,
    ): 
        """
        
        :param nombre:
        :param apellido_paterno:
        :param apellido_materno:
        :param pais_nacionalidad:
        :param actividad_economica:
        :param fecha_nacimiento:
        :param rfc:
        :param curp:
        :param representante_apoderado:
        """
        
        super().__init__({
            'Nombre': nombre,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'PaisNacionalidad': pais_nacionalidad,
            'ActividadEconomica': actividad_economica,
            'FechaNacimiento': fecha_nacimiento,
            'Rfc': rfc,
            'Curp': curp,
            'RepresentanteApoderado': representante_apoderado,
        })
        

class TipoPersonaType(XElement):
    def __init__(
            self,
            persona_fisica: PersonaFisicaType | dict = None,
            persona_moral: PersonaMoralType | dict = None,
            fideicomiso: FideicomisoType | dict = None,
    ): 
        """
        
        :param persona_fisica:
        :param persona_moral:
        :param fideicomiso:
        """
        
        super().__init__({
            'PersonaFisica': persona_fisica,
            'PersonaMoral': persona_moral,
            'Fideicomiso': fideicomiso,
        })
        

class PersonaAvisoType(XElement):
    def __init__(
            self,
            tipo_persona: TipoPersonaType | dict,
            tipo_domicilio: TipoDomicilioType | dict = None,
            telefono: TelefonoType | dict = None,
    ): 
        """
        
        :param tipo_persona:
        :param tipo_domicilio:
        :param telefono:
        """
        
        super().__init__({
            'TipoPersona': tipo_persona,
            'TipoDomicilio': tipo_domicilio,
            'Telefono': telefono,
        })
        

class AlertaType(XElement):
    def __init__(
            self,
            tipo_alerta: str,
            descripcion_alerta: str = None,
    ): 
        """
        
        :param tipo_alerta:
        :param descripcion_alerta:
        """
        
        super().__init__({
            'TipoAlerta': tipo_alerta,
            'DescripcionAlerta': descripcion_alerta,
        })
        

class ModificatorioType(XElement):
    def __init__(
            self,
            folio_modificacion: str,
            descripcion_modificacion: str,
    ): 
        """
        
        :param folio_modificacion:
        :param descripcion_modificacion:
        """
        
        super().__init__({
            'FolioModificacion': folio_modificacion,
            'DescripcionModificacion': descripcion_modificacion,
        })
        

class AvisoType(XElement):
    def __init__(
            self,
            referencia_aviso: str,
            prioridad: str,
            alerta: AlertaType | dict,
            persona_aviso: PersonaAvisoType | dict | Sequence[PersonaAvisoType | dict],
            detalle_operaciones: DetalleOperacionesType | dict,
            modificatorio: ModificatorioType | dict = None,
            dueno_beneficiario: DuenoBeneficiarioType | dict | Sequence[DuenoBeneficiarioType | dict] = None,
    ): 
        """
        
        :param referencia_aviso:
        :param prioridad:
        :param alerta:
        :param persona_aviso:
        :param detalle_operaciones:
        :param modificatorio:
        :param dueno_beneficiario:
        """
        
        super().__init__({
            'ReferenciaAviso': referencia_aviso,
            'Prioridad': prioridad,
            'Alerta': alerta,
            'PersonaAviso': persona_aviso,
            'DetalleOperaciones': detalle_operaciones,
            'Modificatorio': modificatorio,
            'DuenoBeneficiario': dueno_beneficiario,
        })
        

class OcupacionType(XElement):
    def __init__(
            self,
            tipo_ocupacion: int,
            descripcion_otra_ocupacion: str = None,
    ): 
        """
        
        :param tipo_ocupacion:
        :param descripcion_otra_ocupacion:
        """
        
        super().__init__({
            'TipoOcupacion': tipo_ocupacion,
            'DescripcionOtraOcupacion': descripcion_otra_ocupacion,
        })
        

class SujetoObligadoType(XElement):
    def __init__(
            self,
            clave_sujeto_obligado: str,
            clave_actividad: str,
            clave_entidad_colegiada: str = None,
            ocupacion: OcupacionType | dict = None,
            exento: str = None,
    ): 
        """
        
        :param clave_sujeto_obligado:
        :param clave_actividad:
        :param clave_entidad_colegiada:
        :param ocupacion:
        :param exento:
        """
        
        super().__init__({
            'ClaveSujetoObligado': clave_sujeto_obligado,
            'ClaveActividad': clave_actividad,
            'ClaveEntidadColegiada': clave_entidad_colegiada,
            'Ocupacion': ocupacion,
            'Exento': exento,
        })
        

class InformeType(XElement):
    def __init__(
            self,
            mes_reportado: str,
            sujeto_obligado: SujetoObligadoType | dict,
            aviso: AvisoType | dict | Sequence[AvisoType | dict] = None,
    ): 
        """
        
        :param mes_reportado:
        :param sujeto_obligado:
        :param aviso:
        """
        
        super().__init__({
            'MesReportado': mes_reportado,
            'SujetoObligado': sujeto_obligado,
            'Aviso': aviso,
        })
        

class ArchivoType(XElement):
    def __init__(
            self,
            informe: InformeType | dict | Sequence[InformeType | dict],
    ): 
        """
        
        :param informe:
        """
        
        super().__init__({
            'Informe': informe,
        })
        

class Archivo(ArchivoType):
    tag = '{http://www.uif.shcp.gob.mx/recepcion/spr}archivo'




from .formatters import SatCFDI
from .process import invoices_print, payments_print,  invoices_confirmation_print
from .process import invoices_export, payments_export, retentions_export, payments_retentions_export
from .process import filter_invoices_by, filter_payments_by, filter_retenciones_by, complement_invoices_data, InvoiceType
from .email import EmailManager



from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class ComprExt(XElement):
    """
    www.sat.gob.mx/esquemas/ContabilidadE/1_1/AuxiliarFolios
    Nodo opcional para relacionar el detalle de los comprobantes de origen extranjero relacionados con la transacción. Se considera que se debe identificar, el soporte documental, tanto en la provisión, como en el pago y/o cobro de cada una de las cuentas y subcuentas que se vean afectadas. Se convierte en requerido cuando se cuente con la información.
    """
    def __init__(
            self,
            num_fact_ext: str,
            monto_total: Decimal | int,
            tax_id: str = None,
            met_pago_aux: str = None,
            moneda: str = None,
            tip_camb: Decimal | int = None,
    ): 
        """
        Nodo opcional para relacionar el detalle de los comprobantes de origen extranjero relacionados con la transacción. Se considera que se debe identificar, el soporte documental, tanto en la provisión, como en el pago y/o cobro de cada una de las cuentas y subcuentas que se vean afectadas. Se convierte en requerido cuando se cuente con la información.
        
        :param num_fact_ext: Atributo requerido para expresar la clave numérico o alfanumérico del comprobante de origen extranjero que soporte la operación
        :param monto_total: Atributo requerido para expresar el monto total del comprobante de origen extranjero que soporte la operación.
        :param tax_id: Atributo opcional que sirve para expresar el Identificador del contribuyente extranjero. Se convierte en requerido cuando se cuente con la información
        :param met_pago_aux: Atributo opcional para expresar el método de pago de la operación, de acuerdo al catálogo publicado en la página de internet del SAT. Se convierte en requerido cuando se cuente con la información.
        :param moneda: Atributo opcional para expresar el tipo de moneda utilizado en la transacción, de acuerdo al catálogo publicado en la página de internet del SAT. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
        :param tip_camb: Atributo opcional para expresar el tipo de cambio utilizado de acuerdo al tipo de moneda. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
        """
        
        super().__init__({
            'NumFactExt': num_fact_ext,
            'MontoTotal': monto_total,
            'TaxID': tax_id,
            'MetPagoAux': met_pago_aux,
            'Moneda': moneda,
            'TipCamb': tip_camb,
        })
        

class ComprNalOtr(XElement):
    """
    www.sat.gob.mx/esquemas/ContabilidadE/1_1/AuxiliarFolios
    Nodo opcional para relacionar el detalle de los comprobantes de origen nacional relacionados con la transacción, diferente a CFDI, es decir, CFD y/o CBB. Se considera que se debe identificar, el soporte documental, tanto en la provisión, como en el pago y/o cobro de cada una de las cuentas y subcuentas que se vean afectadas. Se convierte en requerido cuando se cuente con la información.
    """
    def __init__(
            self,
            cfd_cbb_num_fol: int,
            monto_total: Decimal | int,
            rfc: str,
            cfd_cbb_serie: str = None,
            met_pago_aux: str = None,
            moneda: str = None,
            tip_camb: Decimal | int = None,
    ): 
        """
        Nodo opcional para relacionar el detalle de los comprobantes de origen nacional relacionados con la transacción, diferente a CFDI, es decir, CFD y/o CBB. Se considera que se debe identificar, el soporte documental, tanto en la provisión, como en el pago y/o cobro de cada una de las cuentas y subcuentas que se vean afectadas. Se convierte en requerido cuando se cuente con la información.
        
        :param cfd_cbb_num_fol: Atributo requerido para expresar el número de folio del comprobante CFD_CBB que soporte la operación.
        :param monto_total: Atributo requerido para expresar el monto total del CFD y/o CBB que soporte la operación. (Incluye IVA en su caso)
        :param rfc: Atributo requerido para expresar el RFC relacionado con la operación. El RFC al que se hace referencia, es el distinto del contribuyente que envía los datos, es decir, el RFC del tercero vinculado.
        :param cfd_cbb_serie: Atributo opcional para expresar la serie del comprobante CFD_CBB que soporte la operación.
        :param met_pago_aux: Atributo opcional para expresar el método de pago de la operación, de acuerdo al catálogo publicado en la página de internet del SAT. Se convierte en requerido cuando se cuente con la información.
        :param moneda: Atributo opcional para expresar el tipo de moneda utilizado en la transacción, de acuerdo al catálogo publicado en la página de internet del SAT. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
        :param tip_camb: Atributo opcional para expresar el tipo de cambio utilizado de acuerdo al tipo de moneda. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
        """
        
        super().__init__({
            'CFD_CBB_NumFol': cfd_cbb_num_fol,
            'MontoTotal': monto_total,
            'RFC': rfc,
            'CFD_CBB_Serie': cfd_cbb_serie,
            'MetPagoAux': met_pago_aux,
            'Moneda': moneda,
            'TipCamb': tip_camb,
        })
        

class ComprNal(XElement):
    """
    www.sat.gob.mx/esquemas/ContabilidadE/1_1/AuxiliarFolios
    Nodo opcional para relacionar el detalle de los comprobantes de origen nacional relacionados con la transacción. Se considera que se debe identificar, el soporte documental, tanto en la provisión, como en el pago y/o cobro de cada una de las cuentas y subcuentas que se vean afectadas. Se convierte en requerido cuando se cuente con la información.
    """
    def __init__(
            self,
            uuid_cfdi: str,
            monto_total: Decimal | int,
            rfc: str,
            met_pago_aux: str = None,
            moneda: str = None,
            tip_camb: Decimal | int = None,
    ): 
        """
        Nodo opcional para relacionar el detalle de los comprobantes de origen nacional relacionados con la transacción. Se considera que se debe identificar, el soporte documental, tanto en la provisión, como en el pago y/o cobro de cada una de las cuentas y subcuentas que se vean afectadas. Se convierte en requerido cuando se cuente con la información.
        
        :param uuid_cfdi: Atributo requerido para expresar la clave UUID del CFDI soporte de la operación. (36 caracteres)
        :param monto_total: Atributo requerido para expresar el monto total del CFDI que soporte la operación (Incluye IVA en su caso)
        :param rfc: Atributo requerido para expresar el RFC relacionado con la operación. El RFC al que se hace referencia, es el distinto del contribuyente que envía los datos, es decir, el RFC del tercero vinculado.
        :param met_pago_aux: Atributo opcional para expresar el método de pago de la operación, de acuerdo al catálogo publicado en la página de internet del SAT. Se convierte en requerido cuando se cuente con la información.
        :param moneda: Atributo opcional para expresar el tipo de moneda utilizado en la transacción, de acuerdo al catálogo publicado en la página de internet del SAT. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
        :param tip_camb: Atributo opcional para expresar el tipo de cambio utilizado de acuerdo al tipo de moneda. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
        """
        
        super().__init__({
            'UUID_CFDI': uuid_cfdi,
            'MontoTotal': monto_total,
            'RFC': rfc,
            'MetPagoAux': met_pago_aux,
            'Moneda': moneda,
            'TipCamb': tip_camb,
        })
        

class DetAuxFol(XElement):
    """
    www.sat.gob.mx/esquemas/ContabilidadE/1_1/AuxiliarFolios
    Nodo opcional para expresar el detalle de los folios de los comprobantes fiscales que integran la póliza.
    """
    def __init__(
            self,
            num_un_iden_pol: str,
            fecha: date,
            compr_nal: ComprNal | dict | Sequence[ComprNal | dict] = None,
            compr_nal_otr: ComprNalOtr | dict | Sequence[ComprNalOtr | dict] = None,
            compr_ext: ComprExt | dict | Sequence[ComprExt | dict] = None,
    ): 
        """
        Nodo opcional para expresar el detalle de los folios de los comprobantes fiscales que integran la póliza.
        
        :param num_un_iden_pol: Atributo requerido para expresar el número único de identificación de la póliza. El campo deberá contener la clave o nombre utilizado por el contribuyente para diferenciar, el tipo de póliza y el número correspondiente. En un mes ordinario no debe repetirse un mismo número de póliza con la clave o nombre asignado por el contribuyente.
        :param fecha: Atributo requerido para expresar la fecha de registro de la póliza.
        :param compr_nal: Nodo opcional para relacionar el detalle de los comprobantes de origen nacional relacionados con la transacción. Se considera que se debe identificar, el soporte documental, tanto en la provisión, como en el pago y/o cobro de cada una de las cuentas y subcuentas que se vean afectadas. Se convierte en requerido cuando se cuente con la información.
        :param compr_nal_otr: Nodo opcional para relacionar el detalle de los comprobantes de origen nacional relacionados con la transacción, diferente a CFDI, es decir, CFD y/o CBB. Se considera que se debe identificar, el soporte documental, tanto en la provisión, como en el pago y/o cobro de cada una de las cuentas y subcuentas que se vean afectadas. Se convierte en requerido cuando se cuente con la información.
        :param compr_ext: Nodo opcional para relacionar el detalle de los comprobantes de origen extranjero relacionados con la transacción. Se considera que se debe identificar, el soporte documental, tanto en la provisión, como en el pago y/o cobro de cada una de las cuentas y subcuentas que se vean afectadas. Se convierte en requerido cuando se cuente con la información.
        """
        
        super().__init__({
            'NumUnIdenPol': num_un_iden_pol,
            'Fecha': fecha,
            'ComprNal': compr_nal,
            'ComprNalOtr': compr_nal_otr,
            'ComprExt': compr_ext,
        })
        

class RepAuxFol(CFDI):
    """
    Estándar de reporte auxiliar de folios de los comprobantes fiscales de las pólizas que se entrega como parte de las pólizas de la contabilidad electrónica.
    """
    tag = '{www.sat.gob.mx/esquemas/ContabilidadE/1_1/AuxiliarFolios}RepAuxFol'
    version = '1.2'
    
    def __init__(
            self,
            rfc: str,
            mes: str,
            anio: int,
            tipo_solicitud: str,
            num_orden: str = None,
            num_tramite: str = None,
            sello: str = None,
            no_certificado: str = None,
            certificado: str = None,
            det_aux_fol: DetAuxFol | dict | Sequence[DetAuxFol | dict] = None,
    ): 
        """
        Estándar de reporte auxiliar de folios de los comprobantes fiscales de las pólizas que se entrega como parte de las pólizas de la contabilidad electrónica.
        
        :param rfc: Atributo requerido para expresar el RFC del contribuyente que envía los datos
        :param mes: Atributo requerido para expresar el mes en que inicia la vigencia del reporte auxiliar de los folios de los comprobantes fiscales de las pólizas.
        :param anio: Atributo requerido para expresar el año al que inicia la vigencia del reporte auxiliar de los folios de los comprobantes fiscales de las pólizas.
        :param tipo_solicitud: Atributo requerido para expresar el tipo de solicitud del reporte auxiliar de los folios de los comprobantes fiscales de las pólizas. ( AF - Acto de Fiscalización; FC - Fiscalización Compulsa; DE - Devolución; CO - Compensación )
        :param num_orden: Atributo opcional para expresar el número de orden asignado al acto de fiscalización al que hace referencia el reporte auxiliar de los folios de los comprobantes fiscales de las pólizas. Requerido para tipo de solicitud = AF y FC. Se convierte en requerido cuando se cuente con la información.
        :param num_tramite: Atributo opcional para expresar el número de trámite asignado a la solicitud de devolución o compensación al que hace referencia el reporte auxiliar de los folios de los comprobantes fiscales de las pólizas. Requerido para tipo de solicitud = DE o CO. Se convierte en requerido cuando se cuente con la información.
        :param sello: Atributo opcional para contener el sello digital del archivo de contabilidad electrónica. El sello deberá ser expresado cómo una cadena de texto en formato Base 64
        :param no_certificado: Atributo opcional para expresar el número de serie del certificado de sello digital que ampara el archivo de contabilidad electrónica, de acuerdo al acuse correspondiente a 20 posiciones otorgado por el sistema del SAT.
        :param certificado: Atributo opcional que sirve para expresar el certificado de sello digital que ampara al archivo de contabilidad electrónica como texto, en formato base 64.
        :param det_aux_fol: Nodo opcional para expresar el detalle de los folios de los comprobantes fiscales que integran la póliza.
        """
        
        super().__init__({
            'Version': self.version,
            'RFC': rfc,
            'Mes': mes,
            'Anio': anio,
            'TipoSolicitud': tipo_solicitud,
            'NumOrden': num_orden,
            'NumTramite': num_tramite,
            'Sello': sello,
            'NoCertificado': no_certificado,
            'Certificado': certificado,
            'DetAuxFol': det_aux_fol,
        })
        




from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class Arrendamientoenfideicomiso(CFDI):
    """
    Complemento para expresar el arrendamiento de bienes de un periodo o ejercicio determinado (incluye FIBRAS).
    """
    tag = '{http://www.sat.gob.mx/esquemas/retencionpago/1/arrendamientoenfideicomiso}Arrendamientoenfideicomiso'
    version = '1.0'
    
    def __init__(
            self,
            pag_prov_efec_por_fiduc: Decimal | int,
            rendim_fideicom: Decimal | int,
            deducc_corresp: Decimal | int,
            mont_tot_ret: Decimal | int = None,
            mont_res_fisc_dist_fibras: Decimal | int = None,
            mont_otros_concept_distr: Decimal | int = None,
            descr_mont_otros_concept_distr: str = None,
    ): 
        """
        Complemento para expresar el arrendamiento de bienes de un periodo o ejercicio determinado (incluye FIBRAS).
        
        :param pag_prov_efec_por_fiduc: Atributo requerido para expresar el importe del pago efectuado por parte del fiduciario al arrendador de bienes en el periodo
        :param rendim_fideicom: Atributo requerido para expresar el importe de los rendimientos obtenidos en el periodo por el arrendamiento de bienes
        :param deducc_corresp: Atributo requerido para expresar el importe de las deducciones correspondientes al arrendamiento de los bienes durante el periodo
        :param mont_tot_ret: Atributo opcional para expresar el monto total de la retención del arrendamiento de los bienes del periodo
        :param mont_res_fisc_dist_fibras: Atributo opcional para expresar el monto del resultado fiscal distribuido por FIBRAS
        :param mont_otros_concept_distr: Atributo opcional para expresar el monto de otros conceptos distribuidos
        :param descr_mont_otros_concept_distr: Atributo opcional para describir los conceptos distribuidos cuando se señalen otros conceptos.
        """
        
        super().__init__({
            'Version': self.version,
            'PagProvEfecPorFiduc': pag_prov_efec_por_fiduc,
            'RendimFideicom': rendim_fideicom,
            'DeduccCorresp': deducc_corresp,
            'MontTotRet': mont_tot_ret,
            'MontResFiscDistFibras': mont_res_fisc_dist_fibras,
            'MontOtrosConceptDistr': mont_otros_concept_distr,
            'DescrMontOtrosConceptDistr': descr_mont_otros_concept_distr,
        })
        




from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class VehiculoNuvoSemEnajenadoFabAlPerm(XElement):
    """
    http://www.sat.gob.mx/renovacionysustitucionvehiculos
    Datos del vehículo nuevo o seminuevo que enajena el fabricante, ensamblador o distribuidor autorizado al permisionario.
    """
    def __init__(
            self,
            ano: int,
            num_placas: str,
            modelo: str = None,
            rfc: str = None,
    ): 
        """
        Datos del vehículo nuevo o seminuevo que enajena el fabricante, ensamblador o distribuidor autorizado al permisionario.
        
        :param ano: Atributo requerido para la expresión del año ó año modelo del vehículo nuevo o seminuevo que enajena el fabricante, ensamblador o distribuidor autorizado.
        :param num_placas: Atributo requerido para expresar el número de placa metálicas de identificación del servicio público federal o, en su caso, del servicio público de autotransporte de pasajeros urbano o suburbano del vehículo nuevo o seminuevo que enajena el fabricante, ensamblador o distribuidor autorizado.
        :param modelo: Atributo opcional para expresar el modelo del vehículo nuevo o seminuevo que enajena el fabricante, ensamblador o distribuidor autorizado.
        :param rfc: Atributo opcional para la Clave del Registro Federal de Contribuyentes del arrendatario en el caso de que el adquiriente del vehículo nuevo o seminuevo sea una arrendadora financiera.
        """
        
        super().__init__({
            'Año': ano,
            'NumPlacas': num_placas,
            'Modelo': modelo,
            'RFC': rfc,
        })
        

class VehiculoUsadoEnajenadoPermAlFab(XElement):
    """
    http://www.sat.gob.mx/renovacionysustitucionvehiculos
    Datos del vehículo usado que enajena el permisionario a cuenta del precio del vehículo nuevo o seminuevo.
    """
    def __init__(
            self,
            precio_veh_usado: Decimal | int,
            tipo_veh: str,
            marca: str,
            tipo_o_clase: str,
            ano: int,
            num_placas: str,
            num_fol_tarj_cir: str,
            num_fol_aviso_int: str,
            num_ped_im: str,
            aduana: str,
            fecha_regul_veh: date,
            foliofiscal: str,
            modelo: str = None,
            niv: str = None,
            num_serie: str = None,
            num_motor: str = None,
    ): 
        """
        Datos del vehículo usado que enajena el permisionario a cuenta del precio del vehículo nuevo o seminuevo.
        
        :param precio_veh_usado: Atributo requerido que expresa el precio del vehículo usado que el permisionario enajena al fabricante, ensamblador o distribuidor autorizado a cuenta del precio del vehículo nuevo o seminuevo.
        :param tipo_veh: Atributo requerido que expresar, según el Decreto, las características del vehículo usado que el permisionario enajena al fabricante, ensamblador o distribuidor autorizado a cuenta del precio del vehículo nuevo o seminuevo, de acuerdo con el catálogo “4. Tipo de vehículo conforme al Decreto por el que se otorgan medidas para la sustitución de vehículos de autotransporte de pasaje y carga”.
        :param marca: Atributo requerido para expresar la marca del vehículo usado que se enajena.
        :param tipo_o_clase: Atributo requerido para expresar el tipo o clase del vehículo usado que se enajena.
        :param ano: Atributo requerido para la expresión del año ó año modelo del vehículo usado que se enajena.
        :param num_placas: Atributo requerido para expresar el número de placas metálicas de identificación del servicio público federal o, en su caso, del servicio público de autotransporte de pasajeros urbano o suburbano del vehículo usado que se enajena.
        :param num_fol_tarj_cir: Atributo requerido para expresar el número de folio de la tarjeta de circulación del vehículo usado que se enajena.
        :param num_fol_aviso_int: Atributo requerido para expresar el número de folio del acuse de recibo del Aviso de Intención para acceder al programa de destrucción.
        :param num_ped_im: Atributo requerido para expresar el número de documento aduanero con el cual se importó en definitiva el vehículo usado.
        :param aduana: Atributo requerido para precisar la aduana por la que se efectuó la importación del vehículo usado.
        :param fecha_regul_veh: Atributo requerido para expresar la fecha del pedimento en el que se regularizó la legal importación definitiva del vehículo usado. Se expresa en la forma aaaa-mm-dd de acuerdo a a especificación ISO 8601
        :param foliofiscal: Atributo requerido para expresar el número de folio fiscal del CFDI expedido por el Centro de Destrucción Autorizado al que se ha incorporado el Complemento Certificado de Destrucción del vehículo usado que enajena el permisionario.
        :param modelo: Atributo opcional para la expresión del modelo del vehículo usado que se enajena.
        :param niv: Atributo opcional para expresar el número de identificación vehicular del vehículo usado que se enajena. (Cuando exista el NIV deberá incluirse este invariablemente).
        :param num_serie: Atributo opcional para expresar el número de serie de la carrocería del vehículo usado que se enajena. (En caso de contar con dicho número se deberá ingresar convirtiéndose en requerido).
        :param num_motor: Atributo opcional para expresar el número de motor del vehículo usado que se enajena (En caso de contar con dicho número se deberá ingresar volviéndose requerido).
        """
        
        super().__init__({
            'PrecioVehUsado': precio_veh_usado,
            'TipoVeh': tipo_veh,
            'Marca': marca,
            'TipooClase': tipo_o_clase,
            'Año': ano,
            'NumPlacas': num_placas,
            'NumFolTarjCir': num_fol_tarj_cir,
            'NumFolAvisoint': num_fol_aviso_int,
            'NumPedIm': num_ped_im,
            'Aduana': aduana,
            'FechaRegulVeh': fecha_regul_veh,
            'Foliofiscal': foliofiscal,
            'Modelo': modelo,
            'NIV': niv,
            'NumSerie': num_serie,
            'NumMotor': num_motor,
        })
        

class DecretoSustitVehicular(XElement):
    """
    http://www.sat.gob.mx/renovacionysustitucionvehiculos
    Nodo opcional para expresar los datos aplicables al estimulo por la aplicación del Decreto por el que se otorgan medidas para la sustitución de vehículos de autotransporte de pasaje y carga.
    """
    def __init__(
            self,
            veh_enaj: str,
            vehiculo_usado_enajenado_perm_al_fab: VehiculoUsadoEnajenadoPermAlFab | dict,
            vehiculo_nuvo_sem_enajenado_fab_al_perm: VehiculoNuvoSemEnajenadoFabAlPerm | dict,
    ): 
        """
        Nodo opcional para expresar los datos aplicables al estimulo por la aplicación del Decreto por el que se otorgan medidas para la sustitución de vehículos de autotransporte de pasaje y carga.
        
        :param veh_enaj: Atributo requerido para expresar si el vehículo que el fabricante, ensamblador o distribuidor autorizado enajena al permisionario es nuevo o seminuevo, de acuerdo con el catálogo “2. Vehículo enajenado”.
        :param vehiculo_usado_enajenado_perm_al_fab: Datos del vehículo usado que enajena el permisionario a cuenta del precio del vehículo nuevo o seminuevo.
        :param vehiculo_nuvo_sem_enajenado_fab_al_perm: Datos del vehículo nuevo o seminuevo que enajena el fabricante, ensamblador o distribuidor autorizado al permisionario.
        """
        
        super().__init__({
            'VehEnaj': veh_enaj,
            'VehiculoUsadoEnajenadoPermAlFab': vehiculo_usado_enajenado_perm_al_fab,
            'VehiculoNuvoSemEnajenadoFabAlPerm': vehiculo_nuvo_sem_enajenado_fab_al_perm,
        })
        

class VehiculosUsadosEnajenadoPermAlFab(XElement):
    """
    http://www.sat.gob.mx/renovacionysustitucionvehiculos
    Datos del vehículo o vehículos usados que enajena el permisionario a cuenta del precio del vehículo nuevo o seminuevo (pueden enajenarse 1 o más vehículos, por lo que de ser 2 o más se deberán llenar tantos elementos como vehículos usados se enajenen).
    """
    def __init__(
            self,
            precio_veh_usado: Decimal | int,
            tipo_veh: str,
            marca: str,
            tipo_o_clase: str,
            ano: int,
            num_placas: str,
            num_fol_tarj_cir: str,
            foliofiscal: str,
            modelo: str = None,
            niv: str = None,
            num_serie: str = None,
            num_motor: str = None,
            num_ped_im: str = None,
            aduana: str = None,
            fecha_regul_veh: date = None,
    ): 
        """
        Datos del vehículo o vehículos usados que enajena el permisionario a cuenta del precio del vehículo nuevo o seminuevo (pueden enajenarse 1 o más vehículos, por lo que de ser 2 o más se deberán llenar tantos elementos como vehículos usados se enajenen).
        
        :param precio_veh_usado: Atributo requerido que expresa el precio del vehículo usado que el permisionario enajena al fabricante, ensamblador o distribuidor autorizado a cuenta del precio del vehículo nuevo o seminuevo.
        :param tipo_veh: Atributo requerido para expresar, según el Decreto, el tipo de vehículo usado que enajena el permisionario, de acuerdo con el catálogo “3. Tipo de Vehículo conforme al Decreto por el que se fomenta la renovación del parque vehicular del autotransporte”.
        :param marca: Atributo requerido para expresar la marca del vehículo usado que se enajena.
        :param tipo_o_clase: Atributo requerido para expresar el tipo o clase del vehículo usado que se enajena.
        :param ano: Atributo requerido para la expresión del año ó año modelo del vehículo usado que se enajena.
        :param num_placas: Atributo requerido para expresar el número de placas metálicas de identificación del servicio público federal o, en su caso, del servicio público de autotransporte de pasajeros urbano o suburbano del vehículo usado que se enajena.
        :param num_fol_tarj_cir: Atributo requerido para expresar el número de folio de la tarjeta de circulación del vehículo usado que se enajena.
        :param foliofiscal: Atributo requerido para expresar el número de folio fiscal del CFDI expedido por el Centro de Destrucción Autorizado al que se ha incorporado el Complemento Certificado de Destrucción del vehículo usado que enajena el permisionario.
        :param modelo: Atributo opcional para la expresión del modelo del vehículo usado que se enajena.
        :param niv: Atributo opcional para expresar el número de identificación vehicular del vehículo usado que se enajena. (Cuando exista el NIV deberá incluirse este invariablemente).
        :param num_serie: Atributo opcional para expresar el número de serie de la carrocería del vehículo usado que se enajena. (En caso de contar con dicho número se deberá ingresar convirtiéndose en requerido).
        :param num_motor: Atributo opcional para expresar el número de motor del vehículo usado que se enajena (En caso de contar con dicho número se deberá ingresar volviéndose requerido).
        :param num_ped_im: Atributo opcional para expresar el número de documento aduanero con el cual se importó en definitiva el vehículo usado, en su caso.
        :param aduana: Atributo opcional para precisar la aduana por la que se efectuó la importación del vehículo usado, en su caso.
        :param fecha_regul_veh: Atributo opcional para expresar la fecha del pedimento en el que se regularizó la legal importación definitiva del vehículo usado, en su caso. Se expresa en la forma aaaa-mm-dd de acuerdo a especificación ISO 8601.
        """
        
        super().__init__({
            'PrecioVehUsado': precio_veh_usado,
            'TipoVeh': tipo_veh,
            'Marca': marca,
            'TipooClase': tipo_o_clase,
            'Año': ano,
            'NumPlacas': num_placas,
            'NumFolTarjCir': num_fol_tarj_cir,
            'Foliofiscal': foliofiscal,
            'Modelo': modelo,
            'NIV': niv,
            'NumSerie': num_serie,
            'NumMotor': num_motor,
            'NumPedIm': num_ped_im,
            'Aduana': aduana,
            'FechaRegulVeh': fecha_regul_veh,
        })
        

class DecretoRenovVehicular(XElement):
    """
    http://www.sat.gob.mx/renovacionysustitucionvehiculos
    Nodo opcional para expresar los datos aplicables al estimulo por la aplicación del Decreto por el que se fomenta la renovación del parque vehicular del autotransporte.
    """
    def __init__(
            self,
            veh_enaj: str,
            vehiculos_usados_enajenado_perm_al_fab: VehiculosUsadosEnajenadoPermAlFab | dict | Sequence[VehiculosUsadosEnajenadoPermAlFab | dict],
            vehiculo_nuvo_sem_enajenado_fab_al_perm: VehiculoNuvoSemEnajenadoFabAlPerm | dict,
    ): 
        """
        Nodo opcional para expresar los datos aplicables al estimulo por la aplicación del Decreto por el que se fomenta la renovación del parque vehicular del autotransporte.
        
        :param veh_enaj: Atributo requerido para expresar si el vehículo que el fabricante, ensamblador o distribuidor autorizado enajena al permisionario es nuevo o seminuevo, de acuerdo con el catálogo “2. Vehículo enajenado”.
        :param vehiculos_usados_enajenado_perm_al_fab: Datos del vehículo o vehículos usados que enajena el permisionario a cuenta del precio del vehículo nuevo o seminuevo (pueden enajenarse 1 o más vehículos, por lo que de ser 2 o más se deberán llenar tantos elementos como vehículos usados se enajenen).
        :param vehiculo_nuvo_sem_enajenado_fab_al_perm: Datos del vehículo nuevo o seminuevo que enajena el fabricante, ensamblador o distribuidor autorizado al permisionario.
        """
        
        super().__init__({
            'VehEnaj': veh_enaj,
            'VehiculosUsadosEnajenadoPermAlFab': vehiculos_usados_enajenado_perm_al_fab,
            'VehiculoNuvoSemEnajenadoFabAlPerm': vehiculo_nuvo_sem_enajenado_fab_al_perm,
        })
        

class Renovacionysustitucionvehiculos(CFDI):
    """
    Complemento para incorporar la información relativa a los estímulos por la renovación del parque vehicular del autotransporte y por el que se otorgan medidas para la sustitución de vehículos de autotransporte de pasaje y carga.
    """
    tag = '{http://www.sat.gob.mx/renovacionysustitucionvehiculos}renovacionysustitucionvehiculos'
    version = '1.0'
    
    def __init__(
            self,
            tipo_de_decreto: str,
            decreto_renov_vehicular: DecretoRenovVehicular | dict = None,
            decreto_sustit_vehicular: DecretoSustitVehicular | dict = None,
    ): 
        """
        Complemento para incorporar la información relativa a los estímulos por la renovación del parque vehicular del autotransporte y por el que se otorgan medidas para la sustitución de vehículos de autotransporte de pasaje y carga.
        
        :param tipo_de_decreto: Atributo requerido que indica el Decreto de cuya aplicación se trate, de acuerdo con el catálogo “1. Tipo de Decreto”.
        :param decreto_renov_vehicular: Nodo opcional para expresar los datos aplicables al estimulo por la aplicación del Decreto por el que se fomenta la renovación del parque vehicular del autotransporte.
        :param decreto_sustit_vehicular: Nodo opcional para expresar los datos aplicables al estimulo por la aplicación del Decreto por el que se otorgan medidas para la sustitución de vehículos de autotransporte de pasaje y carga.
        """
        
        super().__init__({
            'Version': self.version,
            'TipoDeDecreto': tipo_de_decreto,
            'DecretoRenovVehicular': decreto_renov_vehicular,
            'DecretoSustitVehicular': decreto_sustit_vehicular,
        })
        




from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class TInformacionAduanera(XElement):
    """
    http://www.sat.gob.mx/vehiculousado
    Tipo definido para expresar información aduanera
    """
    def __init__(
            self,
            numero: str,
            fecha: date,
            aduana: str = None,
    ): 
        """
        Tipo definido para expresar información aduanera
        
        :param numero: Atributo requerido para expresar el número del documento aduanero que ampara la importación del bien.
        :param fecha: Atributo requerido para expresar la fecha de expedición del documento aduanero que ampara la importación del bien.
        :param aduana: Atributo opcional para precisar la aduana por la que se efectuó la importación del bien.
        """
        
        super().__init__({
            'Numero': numero,
            'Fecha': fecha,
            'Aduana': aduana,
        })
        

class VehiculoUsado(CFDI):
    """
    Complemento opcional que permite incorporar información a los contribuyentes que enajenen vehículos nuevos a personas físicas que no tributen en los términos de las Secciones I y II del Capítulo II del Título IV de la ley del ISR, y que reciban en contraprestación como resultados de esa enajenación un vehículo usado y dinero
    """
    tag = '{http://www.sat.gob.mx/vehiculousado}VehiculoUsado'
    version = '1.0'
    
    def __init__(
            self,
            monto_adquisicion: Decimal | int,
            monto_enajenacion: Decimal | int,
            clave_vehicular: str,
            marca: str,
            tipo: str,
            modelo: str,
            valor: Decimal | int,
            numero_motor: str = None,
            numero_serie: str = None,
            niv: str = None,
            informacion_aduanera: TInformacionAduanera | dict | Sequence[TInformacionAduanera | dict] = None,
    ): 
        """
        Complemento opcional que permite incorporar información a los contribuyentes que enajenen vehículos nuevos a personas físicas que no tributen en los términos de las Secciones I y II del Capítulo II del Título IV de la ley del ISR, y que reciban en contraprestación como resultados de esa enajenación un vehículo usado y dinero
        
        :param monto_adquisicion: Atributo requerido para expresar el monto de adquisición del vehículo usado según factura original, primera venta
        :param monto_enajenacion: Atributo requerido para expresar el monto de enajenación del vehículo usado
        :param clave_vehicular: Atributo requerido para expresar la clave vehicular del vehículo usado
        :param marca: Atributo requerido para expresar la marca del vehículo usado
        :param tipo: Atributo requerido para expresar el tipo del vehículo usado
        :param modelo: Atributo requerido para expresar el año modelo del vehículo usado
        :param valor: Atributo requerido para expresar el valor del vehículo, establecido en la Guía EBC o Libro Azul (Guía de Información a Comerciantes de Automóviles y Camiones y Aseguradores de la República Mexicana) vigente, emitida por la Asociación Nacional de Comerciantes en Automóviles y Camiones nuevos y usados A.C.
        :param numero_motor: Atributo opcional para expresar el número de motor del vehículo usado (en caso de contar con dicho número se deberá ingresar)
        :param numero_serie: Atributo opcional para expresar el número de serie de la carrocería del vehículo usado (en caso de contar con dicho número se deberá ingresar)
        :param niv: Atributo opcional para expresar el número de identificación vehicular del vehículo usado (Cuando exista el NIV deberá incluirse este invariablemente)
        :param informacion_aduanera: Nodo opcional para introducir la información aduanera aplicable cuando se trate de ventas de primera mano de mercancías importadas.
        """
        
        super().__init__({
            'Version': self.version,
            'MontoAdquisicion': monto_adquisicion,
            'MontoEnajenacion': monto_enajenacion,
            'ClaveVehicular': clave_vehicular,
            'Marca': marca,
            'Tipo': tipo,
            'Modelo': modelo,
            'Valor': valor,
            'NumeroMotor': numero_motor,
            'NumeroSerie': numero_serie,
            'NIV': niv,
            'InformacionAduanera': informacion_aduanera,
        })
        




from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class AcreditamientoIEPS(CFDI):
    """
    Nodo requerido para expresar los detalles de la descripción del concepto para efectos de poder determinar el monto del estímulo aplicable.
    """
    tag = '{http://www.sat.gob.mx/acreditamiento}acreditamientoIEPS'
    version = '1.0'
    
    def __init__(
            self,
            tar: str,
    ): 
        """
        Nodo requerido para expresar los detalles de la descripción del concepto para efectos de poder determinar el monto del estímulo aplicable.
        
        :param tar: Atributo requerido para expresar la clave de la Terminal de Almacenamiento y Reparto (CVE TAR), conforme al catálogo publicado en la página de Internet del SAT, mismo que servirá para identificar la cuota por litro conforme a las tablas que publique la Secretaría de Hacienda y Crédito Público para determinar el monto del estímulo fiscal.
        """
        
        super().__init__({
            'Version': self.version,
            'TAR': tar,
        })
        




import base64
import time

import requests
from lxml import etree

from . import PAC, Environment, Accept, Document, CancelReason, CancelationAcknowledgment, AcceptRejectAcknowledgment, TaxpayerStatus
from .. import ResponseError, CFDI, Signer, XElement, __version__
from ..create.cancela import cancelacionretencion
from ..create.cancela.aceptacionrechazo import SolicitudAceptacionRechazo
from ..create.cancela.cancelacion import Cancelacion


class SWSapien(PAC):
    """
    Luna Soft, S.A. de C.V.
    SW sapien-SmarterWEB
    """
    RFC = "LSO1306189R5"

    def __init__(self, token=None, user=None, password=None, environment=Environment.PRODUCTION):
        super().__init__(environment)
        if token:
            def get_token():
                return "Bearer " + token
        else:
            def get_token():
                if not self.token or time.time() > self.expires_in:
                    auth_obj = self._authenticate(user, password)
                    self.token = auth_obj["token"]
                    self.expires_in = auth_obj["expires_in"]
                return "Bearer " + self.token

        self.token_fn = get_token

    def _host(self, service="services"):
        match service:
            case "services":
                match self.environment:
                    case Environment.PRODUCTION:
                        return "https://services.sw.com.mx"
                    case Environment.TEST:
                        return "https://services.test.sw.com.mx"
            case "api":
                match self.environment:
                    case Environment.PRODUCTION:
                        return "https://api.sw.com.mx"
                    case Environment.TEST:
                        return "https://api.test.sw.com.mx"

        raise NotImplementedError("Environment not Supported")

    def _request(self, path, headers=None, method="post", files=None, json=None, needs_token=True, service="services"):
        host = self._host(service)

        r = requests.request(
            method=method,
            url=f"{host}/{path}",
            headers={
                'Cache-Control': "no-cache",
                "Authorization": self.token_fn() if needs_token else "",
                "User-Agent": __version__.__user_agent__,
                **(headers or {})
            },
            json=json,
            files=files
        )

        if r.ok:
            res = r.json()
            if res["status"] != "success":
                raise ResponseError(r)
            return res
        else:
            raise ResponseError(r)

    def _authenticate(self, user, password):
        return self._request(
            path="/security/authenticate",
            headers={'user': user, 'password': password},
            needs_token=False
        )

    def _generate_pdf(self, xml: str):
        return self._request(
            path="pdf/v1/api/GeneratePdf",
            service="api",
            json={
                "xmlContent": xml,
                # "logo": None,
                # "extras": None,
                # "templateId": None
            }
        )

    def _issue_stamp(self, cfdi: CFDI, accept: Accept = Accept.XML, ref_id: str = None, operation="stamp") -> Document:
        headers = {}
        if ref_id:
            headers["customid"] = ref_id

        res = self._request(
            path=f"v4/cfdi33/{operation}/v4",
            files=[
                ('xml', ('xml', cfdi.xml_bytes(), 'text/xml')),
            ],
            headers=headers
        )

        pdf = None
        if accept & Accept.PDF:
            pdf = self._generate_pdf(xml=res["data"]["cfdi"])
            pdf = base64.b64decode(pdf['data']['contentB64'])

        return Document(
            document_id=res["data"]["uuid"],
            xml=res["data"]["cfdi"].encode() if accept & accept.XML else None,
            pdf=pdf
        )

    def validate(self, cfdi: CFDI):
        res = self._request(
            path=f"validate/cfdi",
            files=[
                ('xml', ('xml', cfdi.xml_bytes(), 'text/xml')),
            ],
        )
        return res

    def issue(self, cfdi: CFDI, accept: Accept = Accept.XML, ref_id: str = None) -> Document:
        return self._issue_stamp(cfdi=cfdi, accept=accept, ref_id=ref_id, operation="issue")

    def stamp(self, cfdi: CFDI, accept: Accept = Accept.XML, ref_id: str = None) -> Document:
        return self._issue_stamp(cfdi=cfdi, accept=accept, ref_id=ref_id, operation="stamp")

    def recover(self, document_id: str, accept: Accept = Accept.XML) -> Document:
        return self._request(
            path=f"datawarehouse/v1/live/{document_id}",
            service="api",
            method="get"
        )

    def cancel(self, cfdi: CFDI, reason: CancelReason, substitution_id: str = None, signer: Signer = None) -> CancelationAcknowledgment:
        document_id = cfdi["Complemento"]["TimbreFiscalDigital"]["UUID"]
        rfc = cfdi["Emisor"]["Rfc"]
        path = f"cfdi33/cancel/{rfc}/{document_id}/{reason.value}"
        if substitution_id:
            path += f"/{substitution_id}"

        res = self._request(
            path=path
        )

        return CancelationAcknowledgment(
            code=res["data"]["uuid"][document_id],
            acuse=res["data"]["acuse"]
        )

    def cancel_comprobante(self, cancelation: Cancelacion) -> CancelationAcknowledgment:
        res = self._request(
            path=f"cfdi33/cancel/xml",
            files=[
                ('xml', ('xml', etree.tostring(cancelation.to_xml(), encoding="UTF-8"), 'text/xml')),
            ]
        )

        return CancelationAcknowledgment(
            code=res["data"]["uuid"],
            acuse=res["data"]["acuse"]
        )

    def cancel_retencion(self, cancelation: cancelacionretencion.Cancelacion) -> CancelationAcknowledgment:
        raise NotImplementedError()

    def relations(self, cfdi: CFDI):
        document_id = cfdi["Complemento"]["TimbreFiscalDigital"]["UUID"]
        rfc = cfdi["Emisor"]["Rfc"]
        path = f"relations/{rfc}/{document_id}"

        res = self._request(
            path=path
        )

        return res

    def accept_reject(self, request: SolicitudAceptacionRechazo) -> AcceptRejectAcknowledgment:
        res = self._request(
            path=f"acceptreject/xml",
            files=[
                ('xml', ('xml', etree.tostring(request.to_xml(), encoding="UTF-8"), 'text/xml')),
            ]
        )

        return AcceptRejectAcknowledgment(
            folios={
                k["uuid"]: {
                    'estatusUUID': k['estatusUUID'],
                    'respuesta': k['respuesta']
                }
                for k in res["data"]["folios"]
            },
            acuse=res["data"]["acuse"]
        )

    def pending(self, rfc: str) -> list[str]:
        res = self._request(
            path=f"pendings/{rfc}",
            method="get"
        )
        return res['data']['uuid']

    def rfc_valid(self, rfc: str | list[str]) -> bool | list[bool]:
        raise NotImplementedError()

    def list_69b(self, rfc: str) -> TaxpayerStatus:
        res = self._request(
            path=f"taxpayers/{rfc}",
            method="get"
        )
        return res



from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class Remanente(XElement):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/1/dividendos
    Nodo opcional que expresa el resultado obtenido de la diferencia entre ingresos y egresos de las personas morales que distribuyan anticipos o rendimientos o sociedades de producción, sociedades y asociaciones civiles.
    """
    def __init__(
            self,
            proporcion_rem: Decimal | int = None,
    ): 
        """
        Nodo opcional que expresa el resultado obtenido de la diferencia entre ingresos y egresos de las personas morales que distribuyan anticipos o rendimientos o sociedades de producción, sociedades y asociaciones civiles.
        
        :param proporcion_rem: Atributo opcional que expresa el porcentaje de participación de sus integrantes o accionistas
        """
        
        super().__init__({
            'ProporcionRem': proporcion_rem,
        })
        

class DividOUtil(XElement):
    """
    http://www.sat.gob.mx/esquemas/retencionpago/1/dividendos
    Nodo opcional que expresa los dividendos o utilidades distribuidas del periodo o ejercicio
    """
    def __init__(
            self,
            cve_tip_div_outil: str,
            mont_isr_acred_ret_mexico: Decimal | int,
            mont_isr_acred_ret_extranjero: Decimal | int,
            tipo_soc_distr_div: str,
            mont_ret_ext_div_ext: Decimal | int = None,
            mont_isr_acred_nal: Decimal | int = None,
            mont_div_acum_nal: Decimal | int = None,
            mont_div_acum_ext: Decimal | int = None,
    ): 
        """
        Nodo opcional que expresa los dividendos o utilidades distribuidas del periodo o ejercicio
        
        :param cve_tip_div_outil: Atributo requerido para expresar la clave del tipo de dividendo o utilidad distribuida de acuerdo al catálogo.
        :param mont_isr_acred_ret_mexico: Atributo requerido para expresar el importe o retención del dividendo o utilidad en territorio nacional
        :param mont_isr_acred_ret_extranjero: Atributo requerido para expresar el importe o retención del dividendo o utilidad en territorio extranjero
        :param tipo_soc_distr_div: Atributo requerido para expresar si el dividendo es distribuido por sociedades nacionales o extranjeras.
        :param mont_ret_ext_div_ext: Atributo opcional para expresar el monto de la retención en el extranjero sobre dividendos del extranjero
        :param mont_isr_acred_nal: Atributo opcional para expresar el monto del ISR acreditable nacional
        :param mont_div_acum_nal: Atributo opcional para expresar el monto del dividendo acumulable nacional
        :param mont_div_acum_ext: Atributo opcional para expresar el monto del dividendo acumulable extranjero
        """
        
        super().__init__({
            'CveTipDivOUtil': cve_tip_div_outil,
            'MontISRAcredRetMexico': mont_isr_acred_ret_mexico,
            'MontISRAcredRetExtranjero': mont_isr_acred_ret_extranjero,
            'TipoSocDistrDiv': tipo_soc_distr_div,
            'MontRetExtDivExt': mont_ret_ext_div_ext,
            'MontISRAcredNal': mont_isr_acred_nal,
            'MontDivAcumNal': mont_div_acum_nal,
            'MontDivAcumExt': mont_div_acum_ext,
        })
        

class Dividendos(CFDI):
    """
    Complemento para expresar el total de ganancias y utilidades generadas por rendimientos en base a inversiones en instrumentos de inversión
    """
    tag = '{http://www.sat.gob.mx/esquemas/retencionpago/1/dividendos}Dividendos'
    version = '1.0'
    
    def __init__(
            self,
            divid_o_util: DividOUtil | dict = None,
            remanente: Remanente | dict = None,
    ): 
        """
        Complemento para expresar el total de ganancias y utilidades generadas por rendimientos en base a inversiones en instrumentos de inversión
        
        :param divid_o_util: Nodo opcional que expresa los dividendos o utilidades distribuidas del periodo o ejercicio
        :param remanente: Nodo opcional que expresa el resultado obtenido de la diferencia entre ingresos y egresos de las personas morales que distribuyan anticipos o rendimientos o sociedades de producción, sociedades y asociaciones civiles.
        """
        
        super().__init__({
            'Version': self.version,
            'DividOUtil': divid_o_util,
            'Remanente': remanente,
        })
        




from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ...cfdi import CFDI, XElement


class DatosTransito(XElement):
    """
    http://www.sat.gob.mx/TuristaPasajeroExtranjero
    Elemento requerido para expresar la información de la operación realizada
    """
    def __init__(
            self,
            via: str,
            tipo_id: str,
            numero_id: str,
            nacionalidad: str,
            empresa_transporte: str,
            id_transporte: str = None,
    ): 
        """
        Elemento requerido para expresar la información de la operación realizada
        
        :param via: Atributo requerido para expresar si es vía “Aérea”, “Marítima” o "Terrestre"
        :param tipo_id: Atributo requerido para la expresión del número de pasaporte.
        :param numero_id: Atributo requerido para expresar el número de identificación (pasaporte, visa, etc.)
        :param nacionalidad: Atributo requerido para expresar la nacionalidad del turista.
        :param empresa_transporte: Atributo requerido para señalar la empresa de transporte que lo ingresa a territorio nacional o lo traslada de salida.
        :param id_transporte: Atributo opcional para expresar el identificador del medio de transporte usado, ejemplo: número de vuelo.
        """
        
        super().__init__({
            'Via': via,
            'TipoId': tipo_id,
            'NumeroId': numero_id,
            'Nacionalidad': nacionalidad,
            'EmpresaTransporte': empresa_transporte,
            'IdTransporte': id_transporte,
        })
        

class TuristaPasajeroExtranjero(CFDI):
    """
    Complemento opcional al Comprobante Fiscal Digital (CFD) y Comprobante Fiscal Digital a través de Internet (CFDI) para el manejo de datos de TuristaPasajeroExtranjero.
    """
    tag = '{http://www.sat.gob.mx/TuristaPasajeroExtranjero}TuristaPasajeroExtranjero'
    version = '1.0'
    
    def __init__(
            self,
            fechade_transito: datetime,
            tipo_transito: str,
            datos_transito: DatosTransito | dict,
    ): 
        """
        Complemento opcional al Comprobante Fiscal Digital (CFD) y Comprobante Fiscal Digital a través de Internet (CFDI) para el manejo de datos de TuristaPasajeroExtranjero.
        
        :param fechade_transito: Atributo requerido para expresar la fecha y hora del Arribo o Salida del medio de transporte utilizado. Se expresa en la forma aaaa-mm-ddThh:mm:ss, de acuerdo con la especificación ISO 8601.
        :param tipo_transito: Atributo requerido para incorporar la operación realizada: Arribo ó Salida.
        :param datos_transito: Elemento requerido para expresar la información de la operación realizada
        """
        
        super().__init__({
            'Version': self.version,
            'FechadeTransito': fechade_transito,
            'TipoTransito': tipo_transito,
            'DatosTransito': datos_transito,
        })
        




import operator
from datetime import date


class DatePeriod:
    # A class from defining a data range to then do date comparisons

    def __init__(self, year: int | None, month: int = None, day: int = None):
        self.year = year
        if year is None and month is not None:
            raise ValueError("Month can't have value if Year doesn't")
        self.month = month
        if month is None and day is not None:
            raise ValueError("Day can't have value if Month doesn't")
        self.day = day

    def __eq__(self, other):
        if isinstance(other, date):
            for s, o in ((self.year, other.year), (self.month, other.month), (self.day, other.day)):
                if s is not None and s != o:
                    return False
            return True

        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, date):
            return not self.__eq__(other)

        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, date):
            for s, o in ((self.year, other.year), (self.month, other.month), (self.day, other.day)):
                if s is not None and s != o:
                    return s > o
            return False

        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, date):
            return self > other or self == other

        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, date):
            for s, o in ((self.year, other.year), (self.month, other.month), (self.day, other.day)):
                if s is not None and s != o:
                    return s < o
            return False

        return NotImplemented

    def __le__(self, other):
        if isinstance(other, date):
            return self < other or self == other

        return NotImplemented

    def __str__(self):
        res = ""
        for n, f in ((self.year, "{:04d}"), (self.month, "-{:02d}"), (self.day, "-{:02d}")):
            if n:
                res += f.format(n)

        return res



from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing import Iterator

from . import pago20
from ..compute import make_impuestos, rounder, make_impuestos_dr, make_impuesto, \
    make_impuestos_dr_parcial
from ...cfdi import CFDI, XElement, get_timezone
from ...create import Issuer
from ...utils import iterate

_impuestos = {
    "ISR": "001",
    "IVA": "002",
    "IEPS": "003",
}


class Impuesto(XElement):
    """
    http://www.sat.gob.mx/cfd/4
    Nodo requerido para la información detallada de un traslado de impuesto específico.
    """

    def __init__(
            self,
            impuesto: str,
            tipo_factor: str,
            tasa_o_cuota: Decimal | int = None,
            importe: Decimal | int = None,
            base: Decimal | int = None,
    ):
        """
        Nodo requerido para la información detallada de un impuesto específico.

        :param base: Atributo requerido para señalar la suma de los atributos Base de los conceptos del impuesto trasladado. No se permiten valores negativos.
        :param impuesto: Atributo requerido para señalar la clave del tipo de impuesto trasladado.
        :param tipo_factor: Atributo requerido para señalar la clave del tipo de factor que se aplica a la base del impuesto.
        :param tasa_o_cuota: Atributo condicional para señalar el valor de la tasa o cuota del impuesto que se traslada por los conceptos amparados en el comprobante.
        :param importe: Atributo condicional para señalar la suma del importe del impuesto trasladado, agrupado por impuesto, TipoFactor y TasaOCuota. No se permiten valores negativos.
        """

        super().__init__({
            'Base': base,
            'Impuesto': _impuestos.get(impuesto, impuesto),
            'TipoFactor': tipo_factor,
            'TasaOCuota': tasa_o_cuota,
            'Importe': importe,
        })

    @classmethod
    def parse(cls, impuesto: str) -> 'Impuesto':
        parts = impuesto.split("|")
        return cls(
            impuesto=parts[0],
            tipo_factor=parts[1],
            tasa_o_cuota=Decimal(parts[2]) if len(parts) > 2 else None
        )


class CfdiRelacionados(XElement):
    """
    http://www.sat.gob.mx/cfd/4
    Nodo opcional para precisar la información de los comprobantes relacionados.
    """

    def __init__(
            self,
            tipo_relacion: str,
            cfdi_relacionado: str | Sequence[str],
    ):
        """
        Nodo opcional para precisar la información de los comprobantes relacionados.

        :param tipo_relacion: Atributo requerido para indicar la clave de la relación que existe entre éste que se está generando y el o los CFDI previos.
        :param cfdi_relacionado: Nodo requerido para precisar la información de los comprobantes relacionados.
        """

        super().__init__({
            'TipoRelacion': tipo_relacion,
            'CfdiRelacionado': cfdi_relacionado,
        })


class InformacionGlobal(XElement):
    """
    http://www.sat.gob.mx/cfd/4
    Nodo condicional para precisar la información relacionada con el comprobante global.
    """

    def __init__(
            self,
            periodicidad: str,
            meses: str,
            ano: int,
    ):
        """
        Nodo condicional para precisar la información relacionada con el comprobante global.

        :param periodicidad: Atributo requerido para expresar el período al que corresponde la información del comprobante global.
        :param meses: Atributo requerido para expresar el mes o los meses al que corresponde la información del comprobante global.
        :param ano: Atributo requerido para expresar el año al que corresponde la información del comprobante global.
        """

        super().__init__({
            'Periodicidad': periodicidad,
            'Meses': meses,
            'Año': ano,
        })


class Parte(XElement):
    """
    http://www.sat.gob.mx/cfd/4
    Nodo opcional para expresar las partes o componentes que integran la totalidad del concepto expresado en el comprobante fiscal digital por Internet.
    """

    def __init__(
            self,
            clave_prod_serv: str,
            cantidad: Decimal | int,
            descripcion: str,
            no_identificacion: str = None,
            unidad: str = None,
            valor_unitario: Decimal | int = None,
            importe: Decimal | int = None,
            informacion_aduanera: str | Sequence[str] = None,
    ):
        """
        Nodo opcional para expresar las partes o componentes que integran la totalidad del concepto expresado en el comprobante fiscal digital por Internet.

        :param clave_prod_serv: Atributo requerido para expresar la clave del producto o del servicio amparado por la presente parte. Es requerido y deben utilizar las claves del catálogo de productos y servicios, cuando los conceptos que registren por sus actividades correspondan con dichos conceptos.
        :param cantidad: Atributo requerido para precisar la cantidad de bienes o servicios del tipo particular definido por la presente parte.
        :param descripcion: Atributo requerido para precisar la descripción del bien o servicio cubierto por la presente parte.
        :param no_identificacion: Atributo opcional para expresar el número de serie, número de parte del bien o identificador del producto o del servicio amparado por la presente parte. Opcionalmente se puede utilizar claves del estándar GTIN.
        :param unidad: Atributo opcional para precisar la unidad de medida propia de la operación del emisor, aplicable para la cantidad expresada en la parte. La unidad debe corresponder con la descripción de la parte.
        :param valor_unitario: Atributo opcional para precisar el valor o precio unitario del bien o servicio cubierto por la presente parte. No se permiten valores negativos.
        :param importe: Atributo opcional para precisar el importe total de los bienes o servicios de la presente parte. Debe ser equivalente al resultado de multiplicar la cantidad por el valor unitario expresado en la parte. No se permiten valores negativos.
        :param informacion_aduanera: Nodo opcional para introducir la información aduanera aplicable cuando se trate de ventas de primera mano de mercancías importadas o se trate de operaciones de comercio exterior con bienes o servicios.
        """

        super().__init__({
            'ClaveProdServ': clave_prod_serv,
            'Cantidad': cantidad,
            'Descripcion': descripcion,
            'NoIdentificacion': no_identificacion,
            'Unidad': unidad,
            'ValorUnitario': valor_unitario,
            'Importe': importe,
            'InformacionAduanera': informacion_aduanera,
        })


class ACuentaTerceros(XElement):
    """
    http://www.sat.gob.mx/cfd/4
    Nodo opcional para registrar información del contribuyente Tercero, a cuenta del que se realiza la operación.
    """

    def __init__(
            self,
            rfc_a_cuenta_terceros: str,
            nombre_a_cuenta_terceros: str,
            regimen_fiscal_a_cuenta_terceros: str,
            domicilio_fiscal_a_cuenta_terceros: str,
    ):
        """
        Nodo opcional para registrar información del contribuyente Tercero, a cuenta del que se realiza la operación.

        :param rfc_a_cuenta_terceros: Atributo requerido para registrar la Clave del Registro Federal de Contribuyentes del contribuyente Tercero, a cuenta del que se realiza la operación.
        :param nombre_a_cuenta_terceros: Atributo requerido para registrar el nombre, denominación o razón social del contribuyente Tercero correspondiente con el Rfc, a cuenta del que se realiza la operación.
        :param regimen_fiscal_a_cuenta_terceros: Atributo requerido para incorporar la clave del régimen del contribuyente Tercero, a cuenta del que se realiza la operación.
        :param domicilio_fiscal_a_cuenta_terceros: Atributo requerido para incorporar el código postal del domicilio fiscal del Tercero, a cuenta del que se realiza la operación.
        """

        super().__init__({
            'RfcACuentaTerceros': rfc_a_cuenta_terceros,
            'NombreACuentaTerceros': nombre_a_cuenta_terceros,
            'RegimenFiscalACuentaTerceros': regimen_fiscal_a_cuenta_terceros,
            'DomicilioFiscalACuentaTerceros': domicilio_fiscal_a_cuenta_terceros,
        })


class Concepto(XElement):
    """
    http://www.sat.gob.mx/cfd/4
    Nodo requerido para registrar la información detallada de un bien o servicio amparado en el comprobante.
    """

    def __init__(
            self,
            clave_prod_serv: str,
            cantidad: Decimal | int,
            clave_unidad: str,
            descripcion: str,
            valor_unitario: Decimal | int,
            objeto_imp: str = None,
            no_identificacion: str = None,
            unidad: str = None,
            descuento: Decimal | int = None,
            a_cuenta_terceros: ACuentaTerceros | dict = None,
            informacion_aduanera: str | Sequence[str] = None,
            cuenta_predial: str | Sequence[str] = None,
            complemento_concepto: CFDI | Sequence[CFDI] = None,
            parte: Parte | Sequence[Parte | dict] = None,
            traslados: Impuesto | str | Sequence[Impuesto | str | dict] = None,
            retenciones: Impuesto | str | Sequence[Impuesto | str | dict] = None,
            _traslados_incluidos: bool = False
    ):
        """
        Nodo requerido para registrar la información detallada de un bien o servicio amparado en el comprobante.

        :param clave_prod_serv: Atributo requerido para expresar la clave del producto o del servicio amparado por el presente concepto. Es requerido y deben utilizar las claves del catálogo de productos y servicios, cuando los conceptos que registren por sus actividades correspondan con dichos conceptos.
        :param cantidad: Atributo requerido para precisar la cantidad de bienes o servicios del tipo particular definido por el presente concepto.
        :param clave_unidad: Atributo requerido para precisar la clave de unidad de medida estandarizada aplicable para la cantidad expresada en el concepto. La unidad debe corresponder con la descripción del concepto.
        :param descripcion: Atributo requerido para precisar la descripción del bien o servicio cubierto por el presente concepto.
        :param valor_unitario: Atributo requerido para precisar el valor o precio unitario del bien o servicio cubierto por el presente concepto.
        :param objeto_imp: Atributo requerido para expresar si la operación comercial es objeto o no de impuesto.
        :param no_identificacion: Atributo opcional para expresar el número de parte, identificador del producto o del servicio, la clave de producto o servicio, SKU o equivalente, propia de la operación del emisor, amparado por el presente concepto. Opcionalmente se puede utilizar claves del estándar GTIN.
        :param unidad: Atributo opcional para precisar la unidad de medida propia de la operación del emisor, aplicable para la cantidad expresada en el concepto. La unidad debe corresponder con la descripción del concepto.
        :param descuento: Atributo opcional para representar el importe de los descuentos aplicables al concepto. No se permiten valores negativos.
        :param a_cuenta_terceros: Nodo opcional para registrar información del contribuyente Tercero, a cuenta del que se realiza la operación.
        :param informacion_aduanera: Nodo opcional para introducir la información aduanera aplicable cuando se trate de ventas de primera mano de mercancías importadas o se trate de operaciones de comercio exterior con bienes o servicios.
        :param cuenta_predial: Nodo opcional para asentar el número de cuenta predial con el que fue registrado el inmueble, en el sistema catastral de la entidad federativa de que trate, o bien para incorporar los datos de identificación del certificado de participación inmobiliaria no amortizable.
        :param complemento_concepto: Nodo opcional donde se incluyen los nodos complementarios de extensión al concepto definidos por el SAT, de acuerdo con las disposiciones particulares para un sector o actividad específica.
        :param parte: Nodo opcional para expresar las partes o componentes que integran la totalidad del concepto expresado en el comprobante fiscal digital por Internet.
        :param traslados: Traslados a aplicar.
        :param retenciones: Retenciones a aplicar.
        :param _traslados_incluidos: si el valor valor_unitario ya incluye traslados.
        """

        super().__init__({
            'ClaveProdServ': clave_prod_serv,
            'Cantidad': cantidad,
            'ClaveUnidad': clave_unidad,
            'Descripcion': descripcion,
            'ValorUnitario': valor_unitario,
            'ObjetoImp': objeto_imp,
            'NoIdentificacion': no_identificacion,
            'Unidad': unidad,
            'Descuento': descuento,
            'ACuentaTerceros': a_cuenta_terceros,
            'InformacionAduanera': informacion_aduanera,
            'CuentaPredial': cuenta_predial,
            'ComplementoConcepto': complemento_concepto,
            'Parte': parte,
            'Impuestos': {
                'Traslados': traslados,
                'Retenciones': retenciones
            },
            '_traslados_incluidos': _traslados_incluidos
        })


class Receptor(XElement):
    """
    http://www.sat.gob.mx/cfd/4
    Nodo requerido para precisar la información del contribuyente receptor del comprobante.
    """

    def __init__(
            self,
            rfc: str,
            nombre: str,
            domicilio_fiscal_receptor: str,
            regimen_fiscal_receptor: str,
            uso_cfdi: str,
            residencia_fiscal: str = None,
            num_reg_id_trib: str = None,
    ):
        """
        Nodo requerido para precisar la información del contribuyente receptor del comprobante.

        :param rfc: Atributo requerido para registrar la Clave del Registro Federal de Contribuyentes correspondiente al contribuyente receptor del comprobante.
        :param nombre: Atributo requerido para registrar el nombre(s), primer apellido, segundo apellido, según corresponda, denominación o razón social del contribuyente, inscrito en el RFC, del receptor del comprobante.
        :param domicilio_fiscal_receptor: Atributo requerido para registrar el código postal del domicilio fiscal del receptor del comprobante.
        :param regimen_fiscal_receptor: Atributo requerido para incorporar la clave del régimen fiscal del contribuyente receptor al que aplicará el efecto fiscal de este comprobante.
        :param uso_cfdi: Atributo requerido para expresar la clave del uso que dará a esta factura el receptor del CFDI.
        :param residencia_fiscal: Atributo condicional para registrar la clave del país de residencia para efectos fiscales del receptor del comprobante, cuando se trate de un extranjero, y que es conforme con la especificación ISO 3166-1 alpha-3. Es requerido cuando se incluya el complemento de comercio exterior o se registre el atributo NumRegIdTrib.
        :param num_reg_id_trib: Atributo condicional para expresar el número de registro de identidad fiscal del receptor cuando sea residente en el extranjero. Es requerido cuando se incluya el complemento de comercio exterior.
        """

        super().__init__({
            'Rfc': rfc,
            'Nombre': nombre,
            'DomicilioFiscalReceptor': domicilio_fiscal_receptor,
            'RegimenFiscalReceptor': regimen_fiscal_receptor,
            'UsoCFDI': uso_cfdi,
            'ResidenciaFiscal': residencia_fiscal,
            'NumRegIdTrib': num_reg_id_trib,
        })


@dataclass
class PagoComprobante:
    comprobante: CFDI
    num_parcialidad: int = None
    imp_saldo_ant: Decimal | int = None
    imp_pagado: Decimal | int = None

    def __post_init__(self):
        if self.num_parcialidad is None and self.imp_saldo_ant is None and self.imp_pagado is None:
            self.num_parcialidad = 1
            self.imp_saldo_ant = self.comprobante['Total']
            self.imp_pagado = self.comprobante['Total']

        if self.imp_pagado > self.imp_saldo_ant:
            raise ValueError('Importe Pagado debe de ser menor o igual al Importe Saldo Anterior')


def _make_conceptos(conceptos, rnd_fn):
    def make_concepto(concepto):
        trasladados = [x if isinstance(x, Impuesto) else Impuesto.parse(x) for x in iterate(concepto.get("Impuestos", {}).get("Traslados"))]
        retenciones = [x if isinstance(x, Impuesto) else Impuesto.parse(x) for x in iterate(concepto.get("Impuestos", {}).get("Retenciones"))]

        if concepto.get('_traslados_incluidos'):
            s_tasa = sum(c["TasaOCuota"] for c in trasladados if c["TipoFactor"] == "Tasa")
            s_cuota = sum(c["TasaOCuota"] for c in trasladados if c["TipoFactor"] == "Cuota")
            if any(c for c in trasladados if c["TipoFactor"] in ('Tasa', 'Cuota') and (c.get('Base') is not None or c.get('Importe') is not None)):
                raise ValueError("Not possible to compute '_traslados_incluidos' if any 'trasladados' contains 'Base' or 'Importe'")

            valor_unitario = concepto['ValorUnitario']
            valor_unitario = (valor_unitario - s_cuota) / (s_tasa + 1)
            concepto['ValorUnitario'] = rnd_fn(valor_unitario)
        else:
            valor_unitario = concepto['ValorUnitario']

        importe = concepto["Cantidad"] * valor_unitario
        concepto["Importe"] = rnd_fn(importe)

        if concepto.get("ObjetoImp") in ("01", "03"):
            concepto['Impuestos'] = None
        else:
            base = importe - (concepto.get("Descuento") or 0)
            impuestos = {
                imp_t: [
                    make_impuesto(i, base=base, rnd_fn=rnd_fn) for i in imp
                ]
                for imp_t, imp in [('Traslados', trasladados), ('Retenciones', retenciones)] if imp
            }
            concepto['Impuestos'] = impuestos or None
            concepto["ObjetoImp"] = "02" if impuestos else "01"

        return concepto

    return [make_concepto(c) for c in iterate(conceptos)]


# MAIN #
class Comprobante(CFDI):
    """
    Estándar de Comprobante Fiscal Digital por Internet.
    """
    tag = '{http://www.sat.gob.mx/cfd/4}Comprobante'
    version = '4.0'
    complemento_pago = pago20.Pagos

    def __init__(
            self,
            emisor: Issuer,
            lugar_expedicion: str,
            receptor: Receptor | dict,
            conceptos: Concepto | Sequence[Concepto | dict],
            moneda: str = "MXN",
            tipo_de_comprobante: str = "I",
            exportacion: str = "01",
            serie: str = None,
            folio: str = None,
            forma_pago: str = None,
            condiciones_de_pago: str = None,
            tipo_cambio: Decimal | int = None,
            metodo_pago: str = None,
            confirmacion: str = None,
            informacion_global: InformacionGlobal | dict = None,
            cfdi_relacionados: CfdiRelacionados | Sequence[CfdiRelacionados | dict] = None,
            complemento: CFDI | Sequence[CFDI] = None,
            addenda: CFDI | Sequence[CFDI] = None,
            fecha: datetime = None,
    ):
        """
        Estándar de Comprobante Fiscal Digital por Internet.

        :param emisor: Nodo requerido para expresar la información del contribuyente emisor del comprobante.
        :param lugar_expedicion: Atributo requerido para incorporar el código postal del lugar de expedición del comprobante (domicilio de la matriz o de la sucursal).
        :param receptor: Nodo requerido para precisar la información del contribuyente receptor del comprobante.
        :param conceptos: Nodo requerido para listar los conceptos cubiertos por el comprobante.
        :param moneda: Atributo requerido para identificar la clave de la moneda utilizada para expresar los montos, cuando se usa moneda nacional se registra MXN. Conforme con la especificación ISO 4217.
        :param tipo_de_comprobante: Atributo requerido para expresar la clave del efecto del comprobante fiscal para el contribuyente emisor.
        :param exportacion: Atributo requerido para expresar si el comprobante ampara una operación de exportación.
        :param serie: Atributo opcional para precisar la serie para control interno del contribuyente. Este atributo acepta una cadena de caracteres.
        :param folio: Atributo opcional para control interno del contribuyente que expresa el folio del comprobante, acepta una cadena de caracteres.
        :param forma_pago: Atributo condicional para expresar la clave de la forma de pago de los bienes o servicios amparados por el comprobante.
        :param condiciones_de_pago: Atributo condicional para expresar las condiciones comerciales aplicables para el pago del comprobante fiscal digital por Internet. Este atributo puede ser condicionado mediante atributos o complementos.
        :param tipo_cambio: Atributo condicional para representar el tipo de cambio FIX conforme con la moneda usada. Es requerido cuando la clave de moneda es distinta de MXN y de XXX. El valor debe reflejar el número de pesos mexicanos que equivalen a una unidad de la divisa señalada en el atributo moneda. Si el valor está fuera del porcentaje aplicable a la moneda tomado del catálogo c_Moneda, el emisor debe obtener del PAC que vaya a timbrar el CFDI, de manera no automática, una clave de confirmación para ratificar que el valor es correcto e integrar dicha clave en el atributo Confirmacion.
        :param metodo_pago: Atributo condicional para precisar la clave del método de pago que aplica para este comprobante fiscal digital por Internet, conforme al Artículo 29-A fracción VII incisos a y b del CFF.
        :param confirmacion: Atributo condicional para registrar la clave de confirmación que entregue el PAC para expedir el comprobante con importes grandes, con un tipo de cambio fuera del rango establecido o con ambos casos. Es requerido cuando se registra un tipo de cambio o un total fuera del rango establecido.
        :param informacion_global: Nodo condicional para precisar la información relacionada con el comprobante global.
        :param cfdi_relacionados: Nodo opcional para precisar la información de los comprobantes relacionados.
        :param complemento: Nodo opcional donde se incluye el complemento Timbre Fiscal Digital de manera obligatoria y los nodos complementarios determinados por el SAT, de acuerdo con las disposiciones particulares para un sector o actividad específica.
        :param addenda: Nodo opcional para recibir las extensiones al presente formato que sean de utilidad al contribuyente. Para las reglas de uso del mismo, referirse al formato origen.
        :param fecha: Atributo requerido para la expresión de la fecha y hora de expedición del Comprobante Fiscal Digital por Internet. Se expresa en la forma AAAA-MM-DDThh:mm:ss y debe corresponder con la hora local donde se expide el comprobante.
        :return: objeto CFDI
        """
        conceptos = _make_conceptos(conceptos, rnd_fn=rounder(moneda))
        sub_total = sum(c['Importe'] for c in conceptos)
        descuento = sum(c.get('Descuento') or 0 for c in conceptos)
        impuestos = make_impuestos(conceptos)

        total = sub_total - descuento
        if impuestos:
            total += impuestos.get('TotalImpuestosTrasladados', 0)
            total -= impuestos.get('TotalImpuestosRetenidos', 0)

        descuento = descuento or None
        fecha = fecha or datetime.now(tz=get_timezone(lugar_expedicion)).replace(tzinfo=None)

        super().__init__({
            'Version': self.version,
            'Fecha': fecha,
            'NoCertificado': emisor.certificate_number,
            'Certificado': emisor.signer.certificate_base64() if emisor.signer else '',
            'Sello': '',
            'SubTotal': sub_total,
            'Moneda': moneda,
            'Total': total,
            'TipoDeComprobante': tipo_de_comprobante,
            'Exportacion': exportacion,
            'LugarExpedicion': lugar_expedicion,
            'Serie': serie,
            'Folio': folio,
            'FormaPago': forma_pago,
            'CondicionesDePago': condiciones_de_pago,
            'Descuento': descuento,
            'TipoCambio': tipo_cambio,
            'MetodoPago': metodo_pago,
            'Confirmacion': confirmacion,
            'InformacionGlobal': informacion_global,
            'CfdiRelacionados': cfdi_relacionados,
            'Emisor': {
                "Rfc": emisor.rfc,
                "Nombre": emisor.legal_name,
                "RegimenFiscal": emisor.tax_system
            },
            'Receptor': receptor,
            'Conceptos': conceptos,
            'Impuestos': impuestos,
            'Complemento': complemento,
            'Addenda': addenda,
        })
        if emisor.signer:
            self['Sello'] = emisor.signer.sign_sha256(
                self.cadena_original().encode()
            )

    @classmethod
    def pago(
            cls,
            emisor: Issuer,
            lugar_expedicion: str,
            receptor: Receptor | dict,
            complemento_pago: CFDI,
            cfdi_relacionados: CfdiRelacionados | Sequence[CfdiRelacionados | dict] = None,
            confirmacion: str = None,
            serie: str = None,
            folio: str = None,
            addenda: CFDI | Sequence[CFDI] = None,
            fecha: datetime = None) -> CFDI:
        """
        Estándar de Comprobante Fiscal Digital por Internet de Tipo Pago.

        :param emisor: Nodo requerido para expresar la información del contribuyente emisor del comprobante.
        :param lugar_expedicion: Atributo requerido para incorporar el código postal del lugar de expedición del comprobante (domicilio de la matriz o de la sucursal).
        :param receptor: Nodo requerido para precisar la información del contribuyente receptor del comprobante.
        :param complemento_pago: Pago
        :param serie: Atributo opcional para precisar la serie para control interno del contribuyente. Este atributo acepta una cadena de caracteres.
        :param folio: Atributo opcional para control interno del contribuyente que expresa el folio del comprobante, acepta una cadena de caracteres.
        :param confirmacion: Atributo condicional para registrar la clave de confirmación que entregue el PAC para expedir el comprobante con importes grandes, con un tipo de cambio fuera del rango establecido o con ambos casos. Es requerido cuando se registra un tipo de cambio o un total fuera del rango establecido.
        :param cfdi_relacionados: Nodo opcional para precisar la información de los comprobantes relacionados.
        :param addenda: Nodo opcional para recibir las extensiones al presente formato que sean de utilidad al contribuyente. Para las reglas de uso del mismo, referirse al formato origen.
        :param fecha: Atributo requerido para la expresión de la fecha y hora de expedición del Comprobante Fiscal Digital por Internet. Se expresa en la forma AAAA-MM-DDThh:mm:ss y debe corresponder con la hora local donde se expide el comprobante.
        :return: objeto CFDI
        """
        if cls.version == "3.3":
            receptor["UsoCFDI"] = "P01"
        else:
            receptor["UsoCFDI"] = "CP01"

        return cls(
            emisor=emisor,
            lugar_expedicion=lugar_expedicion,
            receptor=receptor,
            conceptos=Concepto(
                clave_prod_serv='84111506',
                cantidad=1,
                clave_unidad='ACT',
                descripcion='Pago',
                valor_unitario=Decimal(0),
                objeto_imp="01"
            ),
            complemento=complemento_pago,
            serie=serie,
            folio=folio,
            moneda='XXX',
            tipo_de_comprobante='P',
            cfdi_relacionados=cfdi_relacionados,
            confirmacion=confirmacion,
            exportacion="01",
            addenda=addenda,
            fecha=fecha
        )

    @classmethod
    def _pago_tipo_cambio(cls, moneda, tipo_cambio):
        # CRP204: El campo TipoCambioP no debe estar presente cuando el campo Moneda contenga ^MXN$ en el nodo Pago
        if cls.complemento_pago.version == "1.0":
            if moneda == 'MXN' and tipo_cambio == 1:
                tipo_cambio = None
        else:
            if moneda == 'MXN' and tipo_cambio is None:
                tipo_cambio = 1
        return tipo_cambio

    @classmethod
    def pago_comprobantes(
            cls,
            emisor: Issuer,
            lugar_expedicion: str,
            comprobantes: CFDI | PagoComprobante | Sequence[CFDI | PagoComprobante],
            fecha_pago: datetime,
            forma_pago: str,
            tipo_cambio: Decimal | int = None,
            cfdi_relacionados: CfdiRelacionados | Sequence[CfdiRelacionados | dict] = None,
            confirmacion: str = None,
            serie: str = None,
            folio: str = None,
            addenda: CFDI | Sequence[CFDI] = None,
            fecha: datetime = None) -> CFDI:
        """
        Estándar de Comprobante Fiscal Digital por Internet de Tipo Pago. Generado a partir de una lista de Comprobantes
        Se asume que los comprobantes se pagan en su totalidad en una sola exhibición

        :param emisor: Nodo requerido para expresar la información del contribuyente emisor del comprobante.
        :param lugar_expedicion: Atributo requerido para incorporar el código postal del lugar de expedición del comprobante (domicilio de la matriz o de la sucursal).
        :param comprobantes: Sequence[Mapping[str, Any]]a de CFDIS de Comprobante de Ingreso para generar el pago por su monto total
        :param fecha_pago: Atributo requerido para expresar la fecha y hora en la que el beneficiario recibe el pago. Se expresa en la forma aaaa-mm-ddThh:mm:ss, de acuerdo con la especificación ISO 8601.En caso de no contar con la hora se debe registrar 12:00:00.
        :param serie: Atributo opcional para precisar la serie para control interno del contribuyente. Este atributo acepta una cadena de caracteres.
        :param folio: Atributo opcional para control interno del contribuyente que expresa el folio del comprobante, acepta una cadena de caracteres.
        :param forma_pago: Atributo condicional para expresar la clave de la forma de pago de los bienes o servicios amparados por el comprobante.
        :param tipo_cambio: Atributo condicional para representar el tipo de cambio FIX conforme con la moneda usada. Es requerido cuando la clave de moneda es distinta de MXN y de XXX. El valor debe reflejar el número de pesos mexicanos que equivalen a una unidad de la divisa señalada en el atributo moneda. Si el valor está fuera del porcentaje aplicable a la moneda tomado del catálogo c_Moneda, el emisor debe obtener del PAC que vaya a timbrar el CFDI, de manera no automática, una clave de confirmación para ratificar que el valor es correcto e integrar dicha clave en el atributo Confirmacion.
        :param confirmacion: Atributo condicional para registrar la clave de confirmación que entregue el PAC para expedir el comprobante con importes grandes, con un tipo de cambio fuera del rango establecido o con ambos casos. Es requerido cuando se registra un tipo de cambio o un total fuera del rango establecido.
        :param cfdi_relacionados: Nodo opcional para precisar la información de los comprobantes relacionados.
        :param addenda: Nodo opcional para recibir las extensiones al presente formato que sean de utilidad al contribuyente. Para las reglas de uso del mismo, referirse al formato origen.
        :param fecha: Atributo requerido para la expresión de la fecha y hora de expedición del Comprobante Fiscal Digital por Internet. Se expresa en la forma AAAA-MM-DDThh:mm:ss y debe corresponder con la hora local donde se expide el comprobante.
        :return: objeto CFDI
        """
        comprobantes = [c if isinstance(c, PagoComprobante) else PagoComprobante(comprobante=c) for c in iterate(comprobantes)]
        first_cfdi = comprobantes[0].comprobante
        moneda = first_cfdi['Moneda']
        receptor = first_cfdi['Receptor'].copy()
        tipo_cambio = cls._pago_tipo_cambio(moneda, tipo_cambio)

        if not all(
                c.comprobante["Moneda"] == moneda
                and c.comprobante["Emisor"]["Rfc"] == emisor.rfc
                and c.comprobante["Emisor"]["RegimenFiscal"] == emisor.tax_system
                and c.comprobante["Receptor"]["Rfc"] == receptor["Rfc"]
                and c.comprobante["Receptor"].get("RegimenFiscalReceptor") == receptor.get("RegimenFiscalReceptor")
                for c in comprobantes
        ):
            raise ValueError("CFDIS are of different RFC's Emisor/Receptor o Moneda")

        return cls.pago(
            emisor=emisor,
            lugar_expedicion=lugar_expedicion,
            receptor=receptor,
            complemento_pago=cls.complemento_pago(
                pago=[{
                    'DoctoRelacionado': [
                        {
                            'IdDocumento': c.comprobante["Complemento"]["TimbreFiscalDigital"]["UUID"],
                            'Serie': c.comprobante.get("Serie"),
                            'Folio': c.comprobante.get("Folio"),
                            'MonedaDR': c.comprobante["Moneda"],
                            'EquivalenciaDR': 1,
                            'MetodoDePagoDR': c.comprobante["MetodoPago"],
                            'NumParcialidad': c.num_parcialidad,
                            'ImpSaldoAnt': c.imp_saldo_ant,
                            'ImpPagado': c.imp_pagado,
                            'ObjetoImpDR': '02' if 'Impuestos' in c.comprobante else '01',
                            'ImpuestosDR': make_impuestos_dr_parcial(
                                conceptos=c.comprobante['Conceptos'],
                                imp_saldo_ant=c.imp_saldo_ant,
                                imp_pagado=c.imp_pagado,
                                total=c.comprobante["Total"],
                                rnd_fn=rounder(c.comprobante["Moneda"])
                            ) if 'Impuestos' in c.comprobante else None
                        } for c in comprobantes
                    ],
                    'FechaPago': fecha_pago,
                    'FormaDePagoP': forma_pago,
                    'MonedaP': moneda,
                    'TipoCambioP': tipo_cambio
                }]
            ),
            cfdi_relacionados=cfdi_relacionados,
            confirmacion=confirmacion,
            serie=serie,
            folio=folio,
            addenda=addenda,
            fecha=fecha
        )

    @classmethod
    def nomina(
            cls,
            emisor: Issuer,
            lugar_expedicion: str,
            receptor: Receptor | dict,
            complemento_nomina: CFDI,
            cfdi_relacionados: CfdiRelacionados | Sequence[CfdiRelacionados | dict] = None,
            confirmacion: str = None,
            serie: str = None,
            folio: str = None,
            addenda: CFDI | Sequence[CFDI] = None,
            fecha: datetime = None) -> CFDI:
        """
        Estándar de Comprobante Fiscal Digital por Internet de Tipo Pago.

        :param emisor: Nodo requerido para expresar la información del contribuyente emisor del comprobante.
        :param lugar_expedicion: Atributo requerido para incorporar el código postal del lugar de expedición del comprobante (domicilio de la matriz o de la sucursal).
        :param receptor: Nodo requerido para precisar la información del contribuyente receptor del comprobante.
        :param complemento_nomina: Pago
        :param serie: Atributo opcional para precisar la serie para control interno del contribuyente. Este atributo acepta una cadena de caracteres.
        :param folio: Atributo opcional para control interno del contribuyente que expresa el folio del comprobante, acepta una cadena de caracteres.
        :param confirmacion: Atributo condicional para registrar la clave de confirmación que entregue el PAC para expedir el comprobante con importes grandes, con un tipo de cambio fuera del rango establecido o con ambos casos. Es requerido cuando se registra un tipo de cambio o un total fuera del rango establecido.
        :param cfdi_relacionados: Nodo opcional para precisar la información de los comprobantes relacionados.
        :param addenda: Nodo opcional para recibir las extensiones al presente formato que sean de utilidad al contribuyente. Para las reglas de uso del mismo, referirse al formato origen.
        :param fecha: Atributo requerido para la expresión de la fecha y hora de expedición del Comprobante Fiscal Digital por Internet. Se expresa en la forma AAAA-MM-DDThh:mm:ss y debe corresponder con la hora local donde se expide el comprobante.
        :return: objeto CFDI
        """
        if cls.version == "3.3":
            receptor["UsoCFDI"] = "P01"
        else:
            receptor["UsoCFDI"] = "CN01"

        concepto = Concepto(
            clave_prod_serv='84111505',
            cantidad=1,
            clave_unidad='ACT',
            descripcion='Pago de nómina',
            valor_unitario=complemento_nomina.get('TotalPercepciones', 0) + complemento_nomina.get('TotalOtrosPagos', 0),
            descuento=complemento_nomina.get('TotalDeducciones'),
            objeto_imp="03"
        )
        return cls(
            emisor=emisor,
            lugar_expedicion=lugar_expedicion,
            receptor=receptor,
            conceptos=concepto,
            complemento=complemento_nomina,
            serie=serie,
            folio=folio,
            moneda='MXN',
            tipo_de_comprobante='N',
            metodo_pago="PUE",
            forma_pago="99",
            cfdi_relacionados=cfdi_relacionados,
            confirmacion=confirmacion,
            exportacion="01",
            addenda=addenda,
            fecha=fecha,
        )
